"""dtactions __init__

including utility functions,
    - getThisDir, for getting the calling directory of module (in site-packages, also when this is a valid symlink),
    - findInSitePackages, supporting getThisDir
    - checkDirectory, check existence of directory, with create=True, do create if directory does not exist yet.
    
Note: -as user, having pipped the package, the scripts run from the site-packages directory,
          no editing in source files is meant to be done
      -as developer, you need to clone the package, then `build_package` and,
       after a `pip uninstall dtactions` do: `flit install --symlink`.
       When you edit a file, either via the site-packages (symlinked) directory or via the cloned directory,
       the changes will come in the cloned file, so can be committed again.
       See more instructions in the file README.md in the source directory of the package.

Start with the following lines near the top of your python file:

```
try:
    from dtactions.__init__ import getThisDir, checkDirectory
except ModuleNotFoundError:
    print(f'Run this module after "build_package" and "flit install --symlink"\n')
    raise
   
thisDir = getThisDir(__file__) # optional: ... , warnings=True)
```

Also retrieve the DtactionsDirectory and DtactionsUserDirectory from here:

```
from Dtactionscore.__init__ import getDtactionsDirectory, getDtactionsUserDirectory

print(f'DtactionsDirectory: {getDtactionsDirectory()})
print(f'DtactionsUserDirectory: {getDtactionsUserDirectory()})
```

NOTE: these two directories can also be obtained via Dtactionsstatus:
```
import natlinkstatus
status = natlinkstatus.DtactionsStatus()
status.getDtactionsDirectory()
status.getDtactionsUserDirectory()
```

"""

## version to be updated when a new release is sent to pypi:
__version__ = '1.5.4'     # september 2022
#__version__ = '1.5.3'    # august 2022
#__version__ = '1.5.0'    # work in progress, released wxpython for python > 3.8, dragonfly is dependency
#             '1.4.2'     # work in progress, sendkeys reasonably ok, unimacro actions and clipboard not yet...
# __version__ = '1.3.5'   # working on path details, with HOME or DICTATIONTOOLBOXUSER
# __version__ = '1.3.3'   # setting the user directory
                          # adding getDtactionsDirectory and getDtactionsUserDirectory
##----------------------
import sys
import os
from pathlib import Path, WindowsPath

def getDtactionsDirectory():
    """return the root directory of dtactions
    """
    return getThisDir(__file__)

def getDtactionsUserDirectory():
    """get the NatlinkUserDirectory
    
    Here are config files, especially .natlink
    
    By default the users home directory is taken. This directory can be overridden by setting
    the environment variable DICTATIONTOOLBOXUSER to an existing directory.
    Restart then the calling program
    """
    # default dtHome:
    dictation_toolbox_user = os.getenv("DICTATIONTOOLBOXUSER")
    if dictation_toolbox_user:
        dtHome = Path(dictation_toolbox_user)
        if not dtHome.is_dir():
            dtHome = WindowsPath.home()
            print(f'dtactions.getDtactionsUserDirectory: environment variable DICTATIONTOOLBOXUSER does not point to a valid directory: "{dictation_toolbox_user}", take "{dtHome}"')
    else:
        dtHome = WindowsPath.home()

    dtactions_ini_folder = dtHome / ".dtactions"
    if not dtactions_ini_folder.is_dir():
        dtactions_ini_folder.mkdir()   #make it if it doesn't exist
    return str(dtactions_ini_folder)

def getThisDir(fileOfModule, warnings=False):
    """get directory of calling module, if possible in site-packages
    
    call at top of module with `getThisDir(__file__)`
    
    If you want to get warnings (each one only once, pass `warnings = True`)
    
    More above and in the explanation of findInSitePackages.
    """
    thisFile = Path(fileOfModule)
    thisDir = thisFile.parent
    thisDir = findInSitePackages(thisDir, warnings=warnings)
    return thisDir

def findInSitePackages(directory, warnings):
    """get corresponding directory in site-packages 
    
    For users, just having pipped this package, the "directory" is returned, assuming it is in
    the site-packages.
    
    For developers, the directory is either
    --in a clone from github.
        The corresponding directory in site-packages should be a symlink,
        otherwise there was no "flit install --symlink" yet.
    --a directory in the site-packages. This directory should be a symlink to a cloned dir.
    
    The site-packages directory is returned, but the actual files accessed are in the cloned directory.
    
    To get this "GOOD" situation, you perform the steps as pointed out above (or in the README.md file)

    When problems arise, set warnings=True, in the call, preferably when calling getThisDir in the calling module.
    """
    dirstr = str(directory)
    if dirstr.find('\\src\\') < 0:
        if warnings:
            warning(f'directory {dirstr} not connected to a github clone directory, changes will not persist across updates...')
        return directory

    commonpart = dirstr.rsplit('\\src\\', maxsplit=1)[-1]
    spDir = Path(sys.prefix, 'Lib', 'site-packages', commonpart)
    if spDir.is_dir():
        spResolve = spDir.resolve()
        if spResolve == spDir:
            if warnings:
                warning(f'corresponding site-packages directory is not a symlink: {spDir}.\nPlease use "flit install --symlink" when you want to test this package')
        elif spResolve == directory:
            # print(f'directory is symlink: {spDir} and resolves to {directory} all right')
            ## return the symbolic link in the site-packages directory, that resolves to directory!!
            return spDir
        else:
            if warnings:
                warning(f'directory is symlink: {spDir} but does NOT resolve to {directory}, but to {spResolve}')
    else:
        if warnings:
            warning('findInSitePackages, not a valid directory in site-packages, no "flit install --symlink" yet: {spDir}')
    return directory        

def checkDirectory(directory, create=None):
    """check existence of directory path
    
    create == False, None (default): raise OSError if directory is not there

    create == True: create if not existent yet... 
              raise OSError if something strange happens...
    returns None
    """
    if not isinstance(directory, Path):
        directory = Path(directory)
    if directory.is_dir():
        return
    if create is False:
        raise OSError(f'Cannot find directory {directory}, but it should be there.')
    if directory.exists():
        raise OSError(f'path exists, but is not a directory: {directory}')
    directory.mkdir(parents=True)
    if directory.is_dir():
        print('created directory: {directory}')
    else:
        raise OSError(f'did not manage to create directory: {directory}')

warningTexts = []
def warning(text):
    """print warning only once, if warnings is set!
    
    warnings can be set in the calling functions above...
    """
    textForeward = text.replace("\\", "/")
    if textForeward in warningTexts:
        return
    warningTexts.append(textForeward)
    print(text)
    

if __name__ == "__main__":
    print(f'dtactions_user_directory: "{getDtactionsUserDirectory()}"')
    



# This file is part of Dragonfly and Unimacro
# (c) Copyright 2007, 2008 by Christo Butcher
# (c) Copyright 2002, by Quintijn Hoogenboom
# Licensed under the LGPL.
#pylint:disable=R0912
"""Sending keystrokes to the foreground window.

Usage:

At top your module insert:

:code:`from dtactions.sendkeys import sendkeys`

and then in a function:
    
    :code:`sendkeys("keystrokes")`

This module now adopts the Dragonfly :code:`action_key` module,
so `"{alt+w}"` is (in the function) converted to `"a-w"` etc.

(Quintijn Hoogenboom, 2021-04-04)
"""
import re
from dragonfly.actions import action_key, action_text

chord_pattern = re.compile(r'(\{.*?\})')
split_text_from_key = re.compile(r'(\w+)(.*$)')

# more synonyms can be defined here if necessary:
synonym_keys = dict(esc="escape")
def sendkeys(keys):
    """sends keystrokes via dragonfly class :code:`action_key.Keys`
   
Keystrokes following Unimacro/Vocola convention are translated into Dragonfly notation:

:code:`"{shift+right 4}"` is translated to :code:`"s-right:4"`.
   
Multiple keystrokes, pauses before the individual keys and a pause after all the keys can be given in Dragonfly notation:

:code:`"{down/25:12/100}"` presses the `down` key 12 times,
with a pause before each key of 25/100 seconds
and a pause of 1 second (100/100 seconds) after the 12 down presses


Extra notation (Quintijn):

:code:`"!"` at the end of a chord (the keystrokes definition between the braced, :code:`{` and :code:`}`,
triggers the `use_hardware=True` event,
comparable with the SendSystemKeys mechanism of Dragon.

Tested at bottom of this file interactively...
    """
    m = chord_pattern.search(keys)
    if m:
        matches = chord_pattern.split(keys)
        # print(matches)
        for part in matches:
            if not part:
                continue
            if part.startswith('{'):
                part = part[1:-1].lower()  # strip { and } and make lowercase
                use_hardware = False
               
                if part.endswith("!"):
                    use_hardware = True
                    part = part.rstrip("!")
                    part = part.rstrip()
                parts = part.split("+")
                if len(parts) > 1:
                    rest = parts[-1]
                    modifiers = [p[0] for p in parts[:-1]]
                    modifiers = ''.join(modifiers) + "-"
                else:
                    # no modifiers:
                    modifiers = ""
                    rest = part
                m_key = split_text_from_key.match(rest)
                if m_key:
                    key, attr = m_key.groups()
                else:
                    raise ValueError('sendkeys, found no key to press in {part}, total: {keys}')
                key = synonym_keys.get(key, key)
                # being nice for ext... keys, (extend, extleft etc...)
                if key and len(key) > 3 and key.startswith('ext'):
                    key = key[3:]

                if attr.find(' ') == 0:
                    attr = attr.replace(' ', ':', 1)

                key = modifiers + key + attr
                # print(f'sendkeys, key: {key}')        
                action_key.Key(key, use_hardware=use_hardware).execute()
            else:
                # part_keys = ','.join(t for t in part).replace(" ", "space")
                # action_key.Key(part_keys).execute()
                action_text.Text(part).execute()
    else:
        # all_keys = ','.join(t for t in keys).replace(" ", "space")
        # action_key.Key(all_keys).execute()
        action_text.Text(keys).execute()
    
if __name__ == "__main__":
    # sendkeys("{a 3}") #aaa
    
    # try holding down a key:
    # (first selects slow 4 lines, then fast, then goes back to start)
    # sendkeys("{shift:down}{down/25:4}{shift:up/100}")
    # sendkeys("{shift:down}{down:4/100}{shift:up}")
    # sendkeys("{Up:8}{down 8}")  #
    
    # leaves empty:
    # t3 = 'abc{shift+left/50}{del/100}def{shift+left 2/50}ghi{left 4/100}{shift+end/100}{del/100}{backspace 2}'  
    # sendkeys(t3)   # leave

    # When running next 3 lines several times, starting with cursor on the last line (after #),
    # you will see abcde appear several times:
    # sendkeys("a b c d e")  
    # sendkeys("x y z {home}")  
    # sendkeys("{ctrl+a/100}") #
    # sendkeys("{ctrl+end!}{up/50}{end}")
    # sendkeys("x y z ")
    sendkeys("test, test, met komma.")
    sendkeys("{ctrl+extend}test, test,{ctrl+home}{ctrl+end} met komma.")
    ##
    ##



# Module     : winGuiAuto.py
# Synopsis   : Windows GUI automation utilities
# Programmer : Simon Brunning - simon@brunningonline.net
# Date       : 25 June 2003
# Version    : 1.0 
# Copyright  : Released to the public domain. Provided as-is, with no warranty.
# Notes      : Requires Python 2.3, win32all and ctypes 



# Modifications by Tim Couper - tim@tizmoi.net
# 22 Jul 2004
# findControls: deduplicates the list to be returned 
# findControl: handles win32gui.error from initial call to findControls
# getMenuInfo: improved algorithm for calculating no. of items in a menu
# activateMenuItem: improved algorithm for calculating no. of items in a menu
#           
# GLOBALLY corrected spelling: seperator -> separator
#                            : descendent -> descendant
# added findMenuItem
# adapted for NatSpeak/NatLink interaction, Quintijn Hoogenboomhallo
# october 2009. (starting with watsup utilities)
# pylint: disable=C0302, C0115
'''"Windows GUI automation utilities"

The standard pattern of usage of winGuiAuto, originally written by TimCouper,
is in three stages; identify a
window, identify a control in the window, trigger an action on the control.

The end result is always that you wish to have an effect upon some Windows GUI
control.

The first stage is to identify the window within which the control can be
found. To do this, you can use either :code:`findTopWindow` or :code:`findTopWindows`. 

Usually, specifying caption text, the window's class, or both, will be
sufficient to identify the required window. (Note that the window's class is a
Windows API concept)

Testing can be done interactively at the bottom of the file. You need the hndle of the application in order to
run most of the test functions

Helper class and functions
==========================

'''

import array
import ctypes
import types
import os
import struct
import time
import pprint
import re
## python 3 only:
from urllib.parse import unquote

import pywintypes
import win32api
import win32con
import winxpgui as win32gui

GetWindowText = ctypes.windll.user32.GetWindowTextW
GetWindowTextLength = ctypes.windll.user32.GetWindowTextLengthW

# for SendKeys:
braceExact = re.compile (r'[{][^}]+[}]$')
hasBraces = re.compile (r'([{].+?[}])')
BracesExtractKey = re.compile (r'^[{]((alt|ctrl|shift)[+])*(?P<k>[^ ]+?)( [0-9]+)?[}]$', re.I)

## Scintilla constants:
# from dragonfly.actions import scintillacon

def findTopWindow(wantedText=None,
                  wantedClass=None):
    '''Find the hwnd of a top level window.
    
    You can identify windows using caption (part of) or class,

    Arguments:
      wantedText     Text which the required window's caption must
                     *contain*. (Case insensitive match.)
      wantedClass    Class to which the required window must belong.
    Returns:
      handle if found (the first match)
      None if not found
    '''
    if not (wantedText or wantedClass):
        raise ValueError("findTopWindow, wantedText and/or wantedClass must be passed, not: %s and %s"%
                         (wantedText, wantedClass))
    topWindow = findTopWindows(wantedText, wantedClass, checkImmediate=1)
    return topWindow # checking already done!

# for checking on the fly:
wText = None
wClass = None
checkI = None
selFunc = None

def findTopWindows(wantedText=None, wantedClass=None, selectionFunction=None, checkImmediate=None):
    '''Find the hwnd of top level windows.
    
    You can identify windows using captions, classes, a custom selection
    function, or any combination of these. (Multiple selection criteria are
    ANDed. If this isn't what's wanted, use a selection function.)

    Arguments:
    wantedText          Text which required windows' captions must contain.
    wantedClass         Class to which required windows must belong.
    selectionFunction   Window selection function. Reference to a function
                        should be passed here. The function should take hwnd as
                        an argument, and should return True when passed the
                        hwnd of a desired window.

    Returns:            A list containing the window handles of all top level
                        windows matching the supplied selection criteria.

    Usage example:      optDialogs = findTopWindows(wantedText="Options")
    
    checkImmediate: if true, check the options inside the _windowEnumerationHandler, and
                    quit if criterium has been met (only for wantedText and wantedClass)
    
    '''
    # pylint: disable=W0603, R0912
    global wText, wClass, checkI
    results = []
    topWindows = []
    if checkImmediate:
        checkI, wText, wClass, _selFunc = 1, _normaliseText(wantedText), wantedClass, selectionFunction
    else:
        checkI, wText, wClass, _selFunc = None, None, None, None
        
    for i in range(10):
        try:
            win32gui.EnumWindows(_windowEnumerationHandler, topWindows)
        except win32gui.error:
            print('got error in EnumWindows, try again: %s'% i)
        else:
            break
    if checkI:
        checkI = wText = wClass = None
        if topWindows:
            # return 1 window handle:
            return topWindows[0][0]
        return 0
        
    # checking the results:
    if wantedText:
        wantedText = _normaliseText(wantedText)
    for result in topWindows:
        if isinstance(result, tuple) and len(result) == 3:
            hwnd, windowText, windowClass = result
        else:
            print('invalid entry')
            continue
        if wantedText and \
           _normaliseText(windowText).find(wantedText) == -1:
            continue
        if wantedClass and not windowClass.find(wantedClass) == 0:
            continue
        if selectionFunction and not selectionFunction(hwnd):
            continue
        results.append(hwnd)
    return results
    
def getForegroundWindow():
    """get hndle of the current foreground window
    """
    hndle = win32gui.GetForegroundWindow()
    return hndle
    
def dumpTopWindows(doAll=None):
    '''TODO'''
    defaultTitles = ['', 'Default IME', 'Engine Window']
    defaultClasses = ['MSCTFIME UI']
    topWindows = []
    try:
        win32gui.EnumWindows(_windowEnumerationHandler, topWindows)
    except win32gui.error:
        print('got error enumerating windows')
        
    if doAll:
        return topWindows
    selectedWindows = [(hndle,title,clname) for (hndle,title,clname) in topWindows
        if title not in defaultTitles and clname not in defaultClasses]
    return selectedWindows
    
def dumpWindow(hwnd):
    '''Dump all controls from a window into a nested list
    
    Useful during development, allowing to you discover the structure of the
    contents of a window, showing the text and class of all contained controls.
    
    Think of it as a poor man's Spy++. ;-)

    Arguments:      The window handle of the top level window to dump.

    Returns         A nested list of controls. Each entry consists of the
                    control's hwnd, its text, its class, and its sub-controls,
                    if any.

    Usage example:  replaceDialog = findTopWindow(wantedText='Replace')
                    pprint.pprint(dumpWindow(replaceDialog))
    '''
    windows = []
    try:
        win32gui.EnumChildWindows(hwnd, _windowEnumerationHandler, windows)
    except win32gui.error:
        # No child windows
        return windows # empty here
    for window in windows:
        childHwnd, _windowText, _windowClass = window
        window_content = dumpWindow(childHwnd)
        if window_content:
            windows.append(window_content)
       
    def dedup(thelist):
        '''De-duplicate deeply nested windows list.'''
        def listContainsSublists(thelist):
            return bool([sublist
                         for sublist in thelist
                         if isinstance(sublist, list)])
        found=[]
        def dodedup(thelist):
            todel = []
            for index, thing in enumerate(thelist):
                if isinstance(thing, list) and listContainsSublists(thing):
                    dodedup(thing)
                else:
                    if thing in found:
                        todel.append(index)
                    else:
                        found.append(thing)
            todel.reverse()
            for todel in todel:
                del thelist[todel]
        dodedup(thelist)
    dedup(windows)
    
    return windows

def findControl(topHwnd,
                wantedText=None,
                wantedClass=None,
                selectionFunction=None,
                maxWait=1,
                retryInterval=0.1):   ###, checkImmediate=None):
    '''Find a control.
    
    You can identify a control within a top level window, using caption, class,
    a custom selection function, or any combination of these. (Multiple
    selection criteria are ANDed. If this isn't what's wanted, use a selection
    function.)
    
    If no control matching the specified selection criteria is found
    immediately, further attempts will be made. The retry interval and maximum
    time to wait for a matching control can be specified.

    Arguments:
    topHwnd             The window handle of the top level window in which the
                        required controls reside.
    wantedText          Text which the required control's captions must contain.
    wantedClass         Class to which the required control must belong.
    selectionFunction   Control selection function. Reference to a function
                        should be passed here. The function should take
                        hwnd, controlText, controlClass as its arguments
                        and should return True when passed the
                        hwnd of the desired control.
    maxWait             The maximum time to wait for a matching control, in
                        seconds.
    retryInterval       How frequently to look for a matching control, in
                        seconds

    Returns:            The window handle of the first control matching the
                        supplied selection criteria.
                    
    Raises:
    WinGuiAutoError     When no control or multiple controls found.

    Usage example:      optDialog = findTopWindow(wantedText="Options")
                        okButton = findControl(optDialog,
                                               wantedClass="Button",
                                               wantedText="OK")
    '''
    # pylint: disable=R0913
    controls = findControls(topHwnd,
                            wantedText=wantedText,
                            wantedClass=wantedClass,
                            selectionFunction=selectionFunction)
    # check for None returned:  Tim 6 Jul 2004
    if controls is None:
        raise WinGuiAutoError("EnumChildWindows failed with win32gui.error "  +
                              repr(topHwnd) +
                              ", wantedText=" +
                              repr(wantedText) +
                              ", wantedClass=" +
                              repr(wantedClass) +
                              ", selectionFunction=" +
                              repr(selectionFunction) +
                              ", maxWait=" +
                              repr(maxWait) +
                              ", retryInterval=" +
                              str(retryInterval) 
                              )
    
    if len(controls) > 1:
        raise WinGuiAutoError("Multiple controls found for topHwnd=" +
                              repr(topHwnd) +
                              ", wantedText=" +
                              repr(wantedText) +
                              ", wantedClass=" +
                              repr(wantedClass) +
                              ", selectionFunction=" +
                              repr(selectionFunction) +
                              ", maxWait=" +
                              repr(maxWait) +
                              ", retryInterval=" +
                              str(retryInterval))
    elif controls:
        return controls[0]
    if (maxWait-retryInterval) >= 0:
        time.sleep(retryInterval)
        try:
            return findControl(topHwnd=topHwnd,
                               wantedText=wantedText,
                               wantedClass=wantedClass,
                               selectionFunction=selectionFunction,
                               maxWait=maxWait-retryInterval,
                               retryInterval=retryInterval)
        except WinGuiAutoError:
            raise WinGuiAutoError("No control found for topHwnd=" +
                                  repr(topHwnd) +
                                  ", wantedText=" +
                                  repr(wantedText) +
                                  ", wantedClass=" +
                                  repr(wantedClass) +
                                  ", selectionFunction=" +
                                  repr(selectionFunction) +
                                  ", maxWait=" +
                                  repr(maxWait) +
                                  ", retryInterval=" +
                                  str(retryInterval))
        
    raise WinGuiAutoError("No control found for topHwnd=" +
                              repr(topHwnd) +
                              ", wantedText=" +
                              repr(wantedText) +
                              ", wantedClass=" +
                              repr(wantedClass) +
                              ", selectionFunction=" +
                              repr(selectionFunction) +
                              ", maxWait=" +
                              repr(maxWait) +
                              ", retryInterval=" +
                              str(retryInterval))

def findControls(topHwnd,
                 wantedText=None,
                 wantedClass=None,
                 selectionFunction=None):
    '''Find controls.
    
    You can identify controls using captions, classes, a custom selection
    function, or any combination of these. (Multiple selection criteria are
    ANDed. If this isn't what's wanted, use a selection function.)

    Arguments:
    topHwnd             The window handle of the top level window in which the
                        required controls reside.
    wantedText          Text which the required controls' captions must contain.
    wantedClass         Class to which the required controls must belong.
    selectionFunction   Control selection function. Reference to a function
                        should be passed here. The function should take hwnd as
                        an argument, and should return True when passed the
                        hwnd of a desired control.

    Returns:            The window handles of the controls matching the
                        supplied selection criteria.    

    Usage example:      optDialog = findTopWindow(wantedText="Options")
                        def findButtons(hwnd, windowText, windowClass):
                            return windowClass == "Button"
                        buttons = findControl(optDialog, wantedText="Button")
    '''
    if wantedText:
        wantedText = _normaliseText(wantedText)
    def searchChildWindows(currentHwnd):
        results = []
        childWindows = []
        try:
            win32gui.EnumChildWindows(currentHwnd,
                                      _windowEnumerationHandler,
                                      childWindows)
        except win32gui.error:
            # This seems to mean that the control *cannot* have child windows,
            # i.e. is not a container.
            return
        for childHwnd, windowText, windowClass in childWindows:
            descendantMatchingHwnds = searchChildWindows(childHwnd)
            if descendantMatchingHwnds:
                results += descendantMatchingHwnds

            if wantedText:
                if not windowText:
                    continue
                #print('wanted: %s, window: %s'% (wantedText, windowText)
                if _normaliseText(windowText).find(wantedText) == -1:
                    continue
            if wantedClass and \
               not windowClass.find(wantedClass) == 0:
                continue

            if selectionFunction and \
               not selectionFunction(childHwnd, windowText, windowClass):
                continue
            results.append(childHwnd)
        return results

    # deduplicate the returned windows:  Tim 6 Jul 2004
    #return searchChildWindows(topHwnd)
    
    hlist=searchChildWindows(topHwnd)
    
    if hlist:
        ## remove duplicates
        return list(set(hlist))
    else:
        return hlist
    
def findAdditionalControls(controls,
                 wantedText=None,
                 wantedClass=None,
                 selectionFunction=None):
    '''Find controls, from a list of already found controls
    
    You can identify controls using captions, classes, a custom selection
    function, or any combination of these. (Multiple selection criteria are
    ANDed. If this isn't what's wanted, use a selection function.)

    Arguments:
    topHwnd             The window handle of the top level window in which the
                        required controls reside.
    wantedText          Text which the required controls' captions must contain.
    wantedClass         Class to which the required controls must belong.
    selectionFunction   Control selection function. Reference to a function
                        should be passed here. The function should take hwnd as
                        an argument, and should return True when passed the
                        hwnd of a desired control.

    Returns:            The window handles of the controls matching the
                        supplied selection criteria.    

    Usage example:      optDialog = findTopWindow(wantedText="Options")
                        def findButtons(hwnd, windowText, windowClass):
                            return windowClass == "Button"
                        buttons = findControl(optDialog, wantedText="Button")
    '''
    if wantedText:
        wantedText = _normaliseText(wantedText)
    def searchChildWindows(currentHwnd):
        results = []
        childWindows = []
        try:
            win32gui.EnumChildWindows(currentHwnd,
                                      _windowEnumerationHandler,
                                      childWindows)
        except win32gui.error:
            # This seems to mean that the control *cannot* have child windows,
            # i.e. is not a container.
            return
        for childHwnd, windowText, windowClass in childWindows:
            descendantMatchingHwnds = searchChildWindows(childHwnd)
            if descendantMatchingHwnds:
                results += descendantMatchingHwnds

            if wantedText:
                if not windowText:
                    continue
                #print('wanted: %s, window: %s'% (wantedText, windowText)
                if _normaliseText(windowText).find(wantedText) == -1:
                    continue
            if wantedClass and \
               not windowClass.find(wantedClass) == 0:
                continue

            if selectionFunction and \
               not selectionFunction(childHwnd, windowText, windowClass):
                continue
            results.append(childHwnd)
        return results
    

    # deduplicate the returned windows:  Tim 6 Jul 2004
    #return searchChildWindows(topHwnd)
    resultsList = []
    for control in controls:
        hlist=searchChildWindows(control)
    
        if hlist:
            # deduplicate the list:
            hdict={}
            for h in hlist:
                hdict[h]=''
            resultsList.append(hdict.keys())

            
def selFuncExplorerAddress(_hndle, text, Class):
    """needed for quick access to explorer CabinetWClass controls
    
    Adres (Dutch) or Address (English) and other languages ???
    """
    if Class == 'ToolbarWindow32':
        if text and text.startswith('Adres: ') or text.startswith('Address: '):
            return True
    return None

## zie ook _folders:

def getwindowtext(hwnd):
    """unicode version of getwindowstext,
    need ctypes, see top of file
    """
    length = GetWindowTextLength(hwnd)
    buff = ctypes.create_unicode_buffer(length + 1)
    GetWindowText(hwnd, buff, length + 1)
    return buff.value
  
def getFolderFromCabinetWClass(hndle):
    """extract folder from a CabinetWClass window
    
    (Unimacro grammar _folders)
    """
    controls = findControls(hndle, selectionFunction=selFuncExplorerAddress)
    if controls:
        hndle = controls[0]
        text = getwindowtext(hndle)
        ## result is probably u"Address: C:\Documents" or so
        folder = extractFolderFromWindowText(text)
        return folder
    return None

def getFolderFromDialog(hndle, className):
    """dialog for open etc from office etc
    if #32770, but can be other dialogs, like the messages window itself!
    
    """
    if className == '#32770':
        controls = list(findControls(hndle, selectionFunction=selFuncExplorerAddress))
        if controls:
            hndle = controls[0]
            text = getwindowtext(hndle)
            folder = extractFolderFromWindowText(text)
            return folder
    return None
        
def extractFolderFromWindowText(text):
    """get folder info from CabinetWClass or #32770 window:
    
    if "location:" is found, get that info, urlparse unquote(d)
    if otherwise : is found (Adress: or Adres: (enx or nld)) return the info after that :
    
    no check of valid directory is done
    """
    if text.find("location:") >= 0:
        # ms-search:
        folder = text.split("location:", 1)[1]
        folder = unquote(folder)
    elif text.find(":") >= 0:
        # Adress: or Adres:
        folder = text.split(": ", 1)[1]
    else:
        # no activefolder info:
        return None
    return folder
    
def getTopMenu(hWnd):
    '''Get a window's main, top level menu.
    
    Arguments:
    hWnd            The window handle of the top level window for which the top
                    level menu is required.

    Returns:        The menu handle of the window's main, top level menu.

    Usage example:  hMenu = getTopMenu(hWnd)
    '''
    return ctypes.windll.user32.GetMenu(ctypes.c_long(hWnd))

def activateMenuItem(hWnd, menuItemPath):
    '''Activate a menu item
    
    Arguments:
    hWnd                The window handle of the top level window whose menu you 
                        wish to activate.
    menuItemPath        The path to the required menu item. This should be a
                        sequence specifying the path through the menu to the
                        required item. Each item in this path can be specified
                        either as an index, or as a menu name.
                    
    Raises:
    WinGuiAutoError     When the requested menu option isn't found.

    Usage example:      activateMenuItem(notepadWindow, ('file', 'open'))
    
                        Which is exactly equivalent to...
                    
                        activateMenuItem(notepadWindow, (0, 1))
    '''
    # pylint: disable=W0104
    # By Axel Kowald (kowald@molgen.mpg.de)
    # Modified by S Brunning to accept strings in addition to indicies.

    # Top level menu    
    hMenu = getTopMenu(hWnd)
    
    # Get top level menu's item count. Is there a better way to do this?
    # Yes .. Tim Couper 22 Jul 2004
    
    hMenuItemCount=win32gui.GetMenuItemCount(hMenu)
    
##    for hMenuItemCount in xrange(256):
##        try:
##            _getMenuInfo(hMenu, hMenuItemCount)
##        except WinGuiAutoError:
##            break
##    hMenuItemCount -= 1
    
    # Walk down submenus
    for submenu in menuItemPath[:-1]:
        try: # submenu is an index
            0 + submenu
            submenuInfo = _getMenuInfo(hMenu, submenu)
            hMenu, hMenuItemCount = submenuInfo.submenu, submenuInfo.itemCount
        except TypeError: # Hopefully, submenu is a menu name
            try:
                _dump, hMenu, hMenuItemCount = _findNamedSubmenu(hMenu,
                                                                hMenuItemCount,
                                                                submenu)
            except WinGuiAutoError:
                raise WinGuiAutoError("Menu path " +
                                      repr(menuItemPath) +
                                      " cannot be found.")
           
    # Get required menu item's ID. (the one at the end).
    menuItem = menuItemPath[-1]
    try: # menuItem is an index
        0 + menuItem
        menuItemID = ctypes.windll.user32.GetMenuItemID(hMenu,
                                                        menuItem)
    except TypeError: # Hopefully, menuItem is a menu name
        try:
            subMenuIndex, _dump1, _dump2 = _findNamedSubmenu(hMenu,
                                        hMenuItemCount,
                                        menuItem)
        except WinGuiAutoError:
            raise WinGuiAutoError("Menu path " +
                                  repr(menuItemPath) +
                                  " cannot be found.")
        menuItemID = ctypes.windll.user32.GetMenuItemID(hMenu, subMenuIndex)

    # Activate    
    win32gui.PostMessage(hWnd, win32con.WM_COMMAND, menuItemID, 0)
    time.sleep(0.05)
##def findMenuItems(hWnd,wantedText):
##    """Finds menu items whose captions contain the text"""
##    hMenu = getTopMenu(hWnd)
##    hMenuItemCount=win32gui.GetMenuItemCount(hMenu)
##    
##    for topItem in xrange(hMenuItemCount):
##            
    
def getMenuInfo(hWnd, menuItemPath):
    '''TODO'''
    # pylint: disable=W0104
    # Top level menu    
    hMenu = getTopMenu(hWnd)
        
    # Get top level menu's item count. Is there a better way to do this?
    # Yes .. Tim Couper 22 Jul 2004
    hMenuItemCount=win32gui.GetMenuItemCount(hMenu)
    
##    for hMenuItemCount in xrange(256):
##        try:
##            _getMenuInfo(hMenu, hMenuItemCount)
##        except WinGuiAutoError:
##            break
##    hMenuItemCount -= 1
    submenuInfo=None 
    
    # Walk down submenus
    for submenu in menuItemPath:
        try: # submenu is an index
            0 + submenu
            submenuInfo = _getMenuInfo(hMenu, submenu)
            hMenu, hMenuItemCount = submenuInfo.submenu, submenuInfo.itemCount
        except TypeError: # Hopefully, submenu is a menu name
            try:
                submenuIndex, new_hMenu, hMenuItemCount = _findNamedSubmenu(hMenu,
                                                                hMenuItemCount,
                                                                submenu)
                submenuInfo = _getMenuInfo(hMenu, submenuIndex)
                hMenu = new_hMenu
            except WinGuiAutoError:
                raise WinGuiAutoError("Menu path " +
                                      repr(menuItemPath) +
                                      " cannot be found.")
    if submenuInfo is None:
        raise WinGuiAutoError("Menu path " +
                              repr(menuItemPath) +
                              " cannot be found. (Null menu path?)")
        
    
    return submenuInfo
    
def _getMenuInfo(hMenu, uIDItem):
    '''Get various info about a menu item.
    
    Arguments:
    hMenu               The menu in which the item is to be found.
    uIDItem             The item's index

    Returns:            Menu item information object. This object is basically
                        a 'bunch'
                        (see http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52308).
                        It will have useful attributes: name, itemCount,
                        submenu, isChecked, isDisabled, isGreyed, and
                        isSeparator
                    
    Raises:
    WinGuiAutoError     When the requested menu option isn't found.       

    Usage example:      submenuInfo = _getMenuInfo(hMenu, submenu)
                        hMenu, hMenuItemCount = submenuInfo.submenu, submenuInfo.itemCount
    '''
    # An object to hold the menu info
    class MenuInfo(Bunch):
        pass
    menuInfo = MenuInfo()

    # Menu state    
    menuState = ctypes.windll.user32.GetMenuState(hMenu,
                                                  uIDItem,
                                                  win32con.MF_BYPOSITION)
    if menuState == -1:
        raise WinGuiAutoError("No such menu item, hMenu=" +
                              str(hMenu) +
                              " uIDItem=" +
                              str(uIDItem))
    
    # Menu name
    menuName = ctypes.c_buffer("\000" * 32)
    ctypes.windll.user32.GetMenuStringA(ctypes.c_int(hMenu),
                                        ctypes.c_int(uIDItem),
                                        menuName, ctypes.c_int(len(menuName)),
                                        win32con.MF_BYPOSITION)
    menuInfo.name = menuName.value

    # Sub menu info
    menuInfo.itemCount = menuState >> 8
    if bool(menuState & win32con.MF_POPUP):
        menuInfo.submenu = ctypes.windll.user32.GetSubMenu(hMenu, uIDItem)
    else:
        menuInfo.submenu = None
        
    menuInfo.isChecked = bool(menuState & win32con.MF_CHECKED)
    menuInfo.isDisabled = bool(menuState & win32con.MF_DISABLED)
    menuInfo.isGreyed = bool(menuState & win32con.MF_GRAYED)
    menuInfo.isSeparator = bool(menuState & win32con.MF_SEPARATOR)
    # ... there are more, but these are the ones I'm interested in thus far.
    
    return menuInfo

def clickButton(hwnd):
    '''Simulates a single mouse click on a button

    Arguments:
    hwnd    Window handle of the required button.

    Usage example:  okButton = findControl(fontDialog,
                                           wantedClass="Button",
                                           wantedText="OK")
                    clickButton(okButton)
    '''
    _sendNotifyMessage(hwnd, win32con.BN_CLICKED)

def clickStatic(hwnd):
    '''Simulates a single mouse click on a static

    Arguments:
    hwnd    Window handle of the required static.

    Usage example:  TODO
    '''
    _sendNotifyMessage(hwnd, win32con.STN_CLICKED)

def doubleClickStatic(hwnd):
    '''Simulates a double mouse click on a static

    Arguments:
    hwnd    Window handle of the required static.

    Usage example:  TODO
    '''
    _sendNotifyMessage(hwnd, win32con.STN_DBLCLK)

def getComboboxItems(hwnd):
    '''Returns the items in a combo box control.

    Arguments:
    hwnd            Window handle for the combo box.

    Returns:        Combo box items.

    Usage example:  fontCombo = findControl(fontDialog, wantedClass="ComboBox")
                    fontComboItems = getComboboxItems(fontCombo)
    '''
    
    return _getMultipleWindowValues(hwnd,
                                    getCountMessage=win32con.CB_GETCOUNT,
                                    getValueMessage=win32con.CB_GETLBTEXT)

def selectComboboxItem(hwnd, item):
    '''Selects a specified item in a Combo box control.

    Arguments:
    hwnd            Window handle of the required combo box.
    item            The reqired item. Either an index, of the text of the
                    required item.

    Usage example:  fontComboItems = getComboboxItems(fontCombo)
                    selectComboboxItem(fontCombo,
                                       random.choice(fontComboItems))
    '''
    try: # item is an index Use this to select
        0 + item
        win32gui.SendMessage(hwnd, win32con.CB_SETCURSEL, item, 0)
        _sendNotifyMessage(hwnd, win32con.CBN_SELCHANGE)
    except TypeError: # Item is a string - find the index, and use that
        items = getComboboxItems(hwnd)
        itemIndex = items.index(item)
        selectComboboxItem(hwnd, itemIndex)

def getListboxItems(hwnd):
    '''Returns the items in a list box control.

    Arguments:
    hwnd            Window handle for the list box.

    Returns:        List box items.

    Usage example:  docType = findControl(newDialog, wantedClass="ListBox")
                    typeListBox = getListboxItems(docType)
    '''
    
    return _getMultipleWindowValues(hwnd,
                                    getCountMessage=win32con.LB_GETCOUNT,
                                    getValueMessage=win32con.LB_GETTEXT)

def selectListboxItem(hwnd, item):
    '''Selects a specified item in a list box control.

    Arguments:
    hwnd            Window handle of the required list box.
    item            The reqired item. Either an index, of the text of the
                    required item.

    Usage example:  docType = findControl(newDialog, wantedClass="ListBox")
                    typeListBox = getListboxItems(docType)
                    
                    # Select a type at random
                    selectListboxItem(docType,
                                      random.randint(0, len(typeListBox)-1))
    '''
    try: # item is an index Use this to select
        0 + item
        win32gui.SendMessage(hwnd, win32con.LB_SETCURSEL, item, 0)
        _sendNotifyMessage(hwnd, win32con.LBN_SELCHANGE)
    except TypeError: # Item is a string - find the index, and use that
        items = getListboxItems(hwnd)
        itemIndex = items.index(item)
        selectListboxItem(hwnd, itemIndex)

#EM_GETFIRSTVISIBLELINE
def getFirstVisibleLine(hwnd, classname=None):
    
    if classname and classname == 'Scintilla':
        getfirstvisibleline = scintillacon.SCI_GETFIRSTVISIBLELINE
    else:
        getfirstvisibleline = win32con.EM_GETFIRSTVISIBLELINE
    
    return win32gui.SendMessage(hwnd, getfirstvisibleline, 0, 0)
                                    
def getEditText(hwnd, visible=False, classname=None):
    '''Returns the text in an edit control.

    Arguments:
    hwnd            Window handle for the edit control.

    Returns         Edit control text lines. (as a list of lines!)

    Usage example:  pprint.pprint(getEditText(editArea))
    
    Note: this is the slow version, use getWindowTextAll for faster access, but
    the end of line characters must be corrected depending on the control.
    
    '''
    if classname and classname == 'Scintilla':
        getline, getlinecount = scintillacon.SCI_GETLINE, scintillacon.SCI_GETLINECOUNT
    else:
        getline, getlinecount = win32con.EM_GETLINE, win32con.EM_GETLINECOUNT
        
    if visible:
        firstLine = getFirstVisibleLine(hwnd)
    else:
        firstLine = 0
    List =  _getMultipleWindowValues(hwnd,
                                    getCountMessage=win32con.EM_GETLINECOUNT,
                                    getValueMessage=win32con.EM_GETLINE,
                                    first=firstLine)
    return List # [l.rstrip('\n\r') for l in List]
    
def setEditText(hwnd, text, append=False, classname=None):
    '''Set an edit control's text.
    
    Arguments:
    hwnd            The edit control's hwnd.
    text            The text to send to the control. This can be a single
                    string, or a sequence of strings. If the latter, each will
                    be become a a separate line in the control.
    append          Should the new text be appended to the existing text?
                    Defaults to False, meaning that any existing text will be
                    replaced. If True, the new text will be appended to the end
                    of the existing text.
                    Note that the first line of the new text will be directly
                    appended to the end of the last line of the existing text.
                    If appending lines of text, you may wish to pass in an
                    empty string as the 1st element of the 'text' argument.

    Usage example:  print("Enter various bits of text."
                    setEditText(editArea, "Hello, again!")
                    setEditText(editArea, "You still there?")
                    setEditText(editArea, ["Here come", "two lines!"])
                    
                    print("Add some..."
                    setEditText(editArea, ["", "And a 3rd one!"], append=True)
    '''

    if type(text) in (types.StringType,):
        pass
    else:
        text = '\r\n'.join(text)
    #
    ## Ensure that text is a list        
    #
    #try:
    #    text + ''
    #    text = [text]
    #except TypeError:
    #    pass

    # Set the current selection range, depending on append flag
    # -1 (append) means clear selection and put at cursor position
    if classname and classname == "Scintilla":
        setsel, replacesel = scintillacon.SCI_SETSEL, scintillacon.SCI_REPLACESEL
    else:
        setsel, replacesel = win32con.EM_SETSEL, win32con.EM_REPLACESEL
    
    if append:
        win32gui.SendMessage(hwnd, setsel, -1, 0)
    else:
        win32gui.SendMessage(hwnd, setsel, 0, -1)
                             
    # Send the text
    win32gui.SendMessage(hwnd, replacesel, True, text)



def replaceEditText(hwnd, text, classname=None):
    '''replace selection of Edit text with text
    
    Arguments:
    hwnd            The edit control's hwnd.
    text            The text to send to the control. This can be a single
                    string, or a sequence of strings. If the latter, each will
                    be become a a separate line in the control.
    '''
    if classname and classname == "Scintilla":
        setsel, replacesel = scintillacon.SCI_SETSEL, scintillacon.SCI_REPLACESEL
    else:
        setsel, replacesel = win32con.EM_SETSEL, win32con.EM_REPLACESEL
    
    if type(text) in (types.StringType,):
        pass
    else:
        text = '\n'.join(text)
    ## Ensure that text is a list        
    #try:
    #    text + ''
    #    text = [text]
    #except TypeError:
    #    pass

    # assume the selection is set (or just end of buffer)
    # Send the text
    win32gui.SendMessage(hwnd, replacesel, True, text)

def appendEditText(hwnd, text, place="end", classname=None):
    '''replace selection of Edit text with text
    
    Arguments:
    hwnd            The edit control's hwnd.
    text            The text to send to the control. This can be a single
                    string, or a sequence of strings. If the latter, each will
                    be become a a separate line in the control.
    place:          "end" at end of buffer, can also be "start"
    '''
    if classname and classname == "Scintilla":
        replacesel = scintillacon.SCI_SETSEL, scintillacon.SCI_REPLACESEL
    else:
        replacesel = win32con.EM_REPLACESEL
    
    if type(text) in (types.StringType,):
        pass
    else:
        text = '\n'.join(text)

    if place == "end":
        current = os.linesep.join(getEditText(hwnd))
        length = len(current)
        setSelection(hwnd, length, length)
    elif place == "start":
        setSelection(hwnd, 0,0)
    else:
        raise ValueError("appendEditText, invalid place parameter: %s"% place)
        
    # Send the text
    win32gui.SendMessage(hwnd, replacesel, True, text)

def getSelection(hndle, classname=None):
    """returns the start and end position of the selection.
    """
    if classname and classname == "Scintilla":
        getsel= scintillacon.SCI_GETSEL
    else:
        getsel = win32con.EM_GETSEL
    
    selInfo = win32gui.SendMessage(hndle, getsel, 0, 0)
    st = win32gui.LOWORD(selInfo)
    en = win32gui.HIWORD(selInfo)
    return st, en

def isVisible(hndle, classname=None):
    """check if window/control is visible
    """
    if classname == 'Scintilla':
        testnum = scintillacon.SCI_GETFIRSTVISIBLELINE
    else:
        return # not know how to do...
        
    v = win32gui.SendMessage(hndle, testnum)
    return v


def setSelection(hndle, lo, hi, classname=None):

    if classname and classname == "Scintilla":
        setsel = scintillacon.SCI_SETSEL
    else:
        setsel = win32con.EM_SETSEL

    win32gui.SendMessage(hndle, setsel, lo, hi)

def getLineNumber(hndle, charPos=-1, classname=None):
    """get the number of the line containing char or the selection (1 based!)
    """
    num = win32gui.SendMessage(hndle, win32con.EM_LINEFROMCHAR, charPos, 0)
    return num + 1

def getNumberOfLines(hndle, classname=None):
    """retrieves the number of lines in the buffer
    
    the total buflen fails, goes better with getRawTextLength (but is not accurate)
    """
    if classname and classname == "Scintilla":
        getlinecount = scintillacon.SCI_GETLINECOUNT 
    else:
        getlinecount = win32con.EM_GETLINECOUNT
    
    numLines = win32gui.SendMessage(hndle, getlinecount, 0, 0)
    return numLines

# length and buffer for receiving 1 line of text. If text length exceeds the
# LINE_BUFFER_LENGTH, it is doubled.
## 'c' naar 'b'  TODOQH
LINE_BUFFER_LENGTH = 256
LINE_TEXT_BUFFER = array.array('b', b' ' * (LINE_BUFFER_LENGTH))
bufferlength  = struct.pack('i', LINE_BUFFER_LENGTH) # This is a C style int.
LINE_TEXT_BUFFER = array.array('b', bufferlength + b' ' * LINE_BUFFER_LENGTH)

def getTextLine(hndle, lineNum, classname=None):
    """get a line of text from buffer, with line number lineNum
    
    Note: the lines are dependent on the width of the control, if it scrolls,
    the number of lines increases.
    Note: lineNum is 0 based!!!
    """
    global LINE_BUFFER_LENGTH, LINE_TEXT_BUFFER
    if classname and classname == "Scintilla":
        getline= scintillacon.SCI_GETLINE
    else:
        getline = win32con.EM_GETLINE
    
    
    
    
    lenReceived =  win32gui.SendMessage(hndle, getline,
                                        lineNum, LINE_TEXT_BUFFER)
    if (LINE_BUFFER_LENGTH - lenReceived) < 10:
        LINE_BUFFER_LENGTH *= 2
        bufferlength  = struct.pack('i', LINE_BUFFER_LENGTH) # This is a C style int.
        LINE_TEXT_BUFFER = array.array('b', bufferlength + b' ' * LINE_BUFFER_LENGTH)
        print('line too small, double to %s'% LINE_BUFFER_LENGTH)
        return getTextLine(hndle, lineNum)

    return LINE_TEXT_BUFFER.tostring()[:lenReceived]

def getRawTextLength(hndle, classname=None):
    """get the one stroke text length, this is with uncorrected \r\n characters
    
    can be used to check if buffer has changed in length...
    (this call takes nearly as much time as the getWindowTextAll call)
    """
    if classname and classname == "Scintilla":
        gettextlength = scintillacon.SCI_GETTEXTLENGTH
    else:
        gettextlength = win32con.WM_GETTEXTLENGTH
    
    return win32gui.SendMessage(hndle, gettextlength, 0, 0)
#

offsetLetter = 0  # ???           
keySynonyms = dict(backspace="back",
                   enter="return")
keySynonyms['del'] = "delete"


## functions with sendkeys seem not to work. Want the hndle of a control, but this is not doable.
## the _sendkeys via ctypes seems to work better. 
def sendKey(hndle, key):
    """sendKey function from messagesfunctions, not in active use
    """
    # pylint: disable=C0321
    kdown = win32con.WM_KEYDOWN
    char = win32con.WM_CHAR
    kup = win32con.WM_KEYUP
    sm = win32gui.SendMessage
    if isinstance(key, str):
        ctrl = alt = shift = 0
        kCode = None
        if key.startswith("{") and key.endswith("}"):
            key = key.strip("{}")
            keys = key.split("+")
            for k in keys:
                if k == 'ctrl': ctrl = 1
                elif k == 'shift': shift = 1
                elif k == 'alt': alt = 1
                else:
                    k = keySynonyms.get(k, k)
                    kCode = getattr(win32con, "VK_%s"%k.upper(),None)
                    if kCode is None:
                        if len(k) == 1:
                            kCode = ord(k.upper())
                        else:
                            raise ValueError("invalid key: {%s"% key[1:])
            if ctrl:  sm(hndle, kdown, win32con.VK_CONTROL, 0)
            if shift: sm(hndle, kdown, win32con.VK_SHIFT, 0)
            if alt:   sm(hndle, kdown, win32con.VK_MENU, 0)
        
            if kCode:
                sm(hndle, kdown, kCode, 0)
                sm(hndle, kup, kCode, 0)
            if alt:   sm(hndle, kup, win32con.VK_MENU, 0)
            if shift: sm(hndle, kup, win32con.VK_SHIFT, 0)
            if ctrl:  sm(hndle, kup, win32con.VK_CONTROL, 0)
            
        else:
            for k in key:
                if ord(k) >= 32:
                    sm(hndle, char, ord(k), 0)  # from space on
                else:
                    sm(hndle, kdown, ord(k), 0) # eg \r and \n
                    sm(hndle, kup, ord(k), 0)
                    
    else:
        raise ValueError("SendKey expects a string, not: %s"% key)

def waitForForegroundWindow(className=None, nWait=50, waitingTime=0.5):
    """wait some time for foreground window to get in front"""
    if isinstance(className, str):
        className = [className]
    elif className is None:
        print('waitForForegroundWindow: no classname given')
        return None
    
    for _ in range(nWait):
        hndle = getForegroundWindow()
        if hndle:
            currentClass = win32gui.GetClassName(hndle)
            if currentClass in className:
                return hndle
        time.sleep(waitingTime)
    return None

def SendKeys(hndle, keysString):
    """send the keys to the window with hndle
    
    This function could be useful when you want to send keystrokes to
    other windows. But it fails, and should be repaired in that case.
    
    The :code:`SendMessage` trick, with windows messages, *does not work*,
    
    The :code:`SendInput` mechanism (as used in Dragonfly,
    and so in the `sendkeys.py` module of this module) is favourable.
 
    Example:
    ::
      hndle = getForegroundWindow()
      SendKeys(hndle, "{shift+a}")   ## fails
      SendKeys(hndle, "abc")         ## works
    
    """
    keysList = splitKeysString(keysString)
    for key in keysList:
        sendKey(hndle,key)
    
def splitKeysString(keysString):
    """split in single keys, either with braces around or single keystrokes
    """
    L = filter(None, hasBraces.split(keysString))
    M = []
    for item in L:
        if braceExact.match(item):
            M.append(item)
        else:
            M.extend(list(item))
    return M
    

def quitProgram(hwnd):
    """send a WM_QUIT message
    """
    #win32gui.DestroyWindow(hwnd)
    win32gui.PostMessage(hwnd,
                         win32con.WM_DESTROY,
                         0,0)

    #_sendNotifyMessage(hwnd, win32con.WM_ACTIVATE)
    #oldHndle = win32gui.GetActiveWindow()
    #win32gui.SetActiveWindow(hwnd)
    #time.sleep(1)
    #
    #win32gui.PostQuitMessage(hwnd)
    #time.sleep(1)
    #win32gui.SetActiveWindow(oldHndle)
    #

# buffer and length for one stroke get window text:
BUFFER_LENGTH = 1024
TEXT_BUFFER = array.array('b', b' ' * (BUFFER_LENGTH))
def getWindowTextAll(hwnd, rawLength=None):
    """get the buffer of a edit control in one stroke
    
    return the buffer and the raw length (uncorrected for \r\n and after text character)
    """
    # pylint: disable=W0603
    global TEXT_BUFFER, BUFFER_LENGTH
    if rawLength is None:
        rawLength = BUFFER_LENGTH 
    if rawLength > BUFFER_LENGTH:
        BUFFER_LENGTH = rawLength + 1000   # take a little extra!
        TEXT_BUFFER = array.array('b', b' ' * BUFFER_LENGTH)
    valueLength = win32gui.SendMessage(hwnd, win32con.WM_GETTEXT,
                                            BUFFER_LENGTH, TEXT_BUFFER)
    if not valueLength:
        return '', 0
    
    if (BUFFER_LENGTH - valueLength) <= 10:
        print('probably too small buffer for this control: %s'% valueLength)
        rawLength = getRawTextLength(hwnd) + 1000
        return getWindowTextAll(hwnd, rawLength=rawLength)
        
    result = TEXT_BUFFER.tostring()[:valueLength]
    return result, valueLength

def _getMultipleWindowValues(hwnd, getCountMessage, getValueMessage, first=0):
    '''
    
    A common pattern in the Win32 API is that in order to retrieve a
    series of values, you use one message to get a count of available
    items, and another to retrieve them. This internal utility function
    performs the common processing for this pattern.

    Arguments:
    hwnd                Window handle for the window for which items should be
                        retrieved.
    getCountMessage     Item count message.
    getValueMessage     Value retrieval message.

    Returns:            Retrieved items.
    '''
    result = []
    
    MAX_VALUE_LENGTH = 256
    bufferlength  = struct.pack('i', MAX_VALUE_LENGTH) # This is a C style int.
    valuecount = win32gui.SendMessage(hwnd, getCountMessage, 0, 0)
    for itemIndex in range(first, valuecount):
        valuebuffer = array.array('c',
                                  bufferlength +
                                  ' ' * (MAX_VALUE_LENGTH - len(bufferlength)))
        valueLength = win32gui.SendMessage(hwnd,
                                           getValueMessage,
                                           itemIndex,
                                           valuebuffer)
        result.append(valuebuffer.tostring()[:valueLength])
    return result

def _windowEnumerationHandler(hwnd, resultList):
    '''win32gui.EnumWindows() callback.
    
    Pass to win32gui.EnumWindows() or win32gui.EnumChildWindows() to
    generate a list of window handle, window text, window class tuples.
    stop enumeration, if checkI(mmediate) is set and contents match 
    '''
    if hwnd == 0:
        print('_windowEnumerationHandler, hwnd == 0, skip')
        return None
    # try:
    _testHndle = win32gui.GetWindow(hwnd, 0)
    # except pywintypes.error:
    #     if details[0] == 1400:
    #         print(f'caught 1400 error, skip this window: {hwnd}'% hwnd)
    #         return None
    #     raise
    if checkI:
        text = className = None
        if wText:
            text = win32gui.GetWindowText(hwnd)
            if wText and _normaliseText(text).find(wText) == -1:
                return
        if wClass:
            className = win32gui.GetClassName(hwnd)
            if wClass and className.find(wClass) != 0:
                return
        if not wText:
            text = win32gui.GetWindowText(hwnd)
        if not wClass:
            className = win32gui.GetClassName(hwnd)
        # got it, append and stop:
        resultList.append(  (hwnd, text, className) )
        return True
    # pass on window title and class name:
    text = win32gui.GetWindowText(hwnd)
    className = win32gui.GetClassName(hwnd)
    #visible = win32gui.SendMessage(hwnd, 2491)
    resultList.append(  (hwnd, text, className) )   
    return None
    
def _buildWinLong(high, low):
    '''Build a windows long parameter from high and low words.
    
    See http://support.microsoft.com/support/kb/articles/q189/1/70.asp
    '''
    # return ((high << 16) | low)
    return int(struct.unpack('>L',
                             struct.pack('>2H',
                                         high,
                                         low)) [0])
        
def _sendNotifyMessage(hwnd, notifyMessage):
    '''Send a notify message to a control.'''
    win32gui.SendMessage(win32gui.GetParent(hwnd),
                         win32con.WM_COMMAND,
                         _buildWinLong(notifyMessage,
                                       win32api.GetWindowLong(hwnd,
                                                              win32con.GWL_ID)),
                         hwnd)

def _normaliseText(controlText):
    '''Remove '&' characters, but keep the case
    
    leave None or '' unchanged
    
    Useful for matching control text.'''
    if controlText and controlText.find("&"):
        return controlText.replace('&', '')
    return controlText

def _findNamedSubmenu(hMenu, hMenuItemCount, submenuName):
    '''Find the index number of a menu's submenu with a specific name.'''
    for submenuIndex in range(hMenuItemCount):
        submenuInfo = _getMenuInfo(hMenu, submenuIndex)
        # submenuInfoName = submenuInfo.name
        if _normaliseText(submenuInfo.name).startswith(_normaliseText(submenuName)):
            return submenuIndex, submenuInfo.submenu, submenuInfo.itemCount
    raise WinGuiAutoError("No submenu found for hMenu=" +
                          repr(hMenu) +
                          ", hMenuItemCount=" +
                          repr(hMenuItemCount) +
                          ", submenuName=" +
                          repr(submenuName))

def _dedup(thelist):
    '''De-duplicate deeply nested list.'''
    found=[]
    def dodedup(thelist):
        for index, thing in enumerate(thelist):
            if isinstance(thing, list):
                dodedup(thing)
            else:
                if thing in found:
                    del thelist[index]
                else:
                    found.append(thing)
                
                              
class Bunch:
    # pylint: disable=R0903
    def __init__(self, **kwds):
        self.__dict__.update(kwds)
        
    def __str__(self):
        state = ["%s=%r" % (attribute, value)
                 for (attribute, value)
                 in self.__dict__.items()]
        return '\n'.join(state)
    
class WinGuiAutoError(Exception):
    pass

def self_test():
    #print("Let's see what top windows we have at the 'mo:")
    #pprint.pprint(dumpTopWindows())
    #x = input('->')
    print("Open and locate Notepad")
    os.startfile('notepad')
    notepadWindow = findTopWindow(wantedClass='Notepad')
    _x = input('->')
    print("Locate the 'find' edit box")
    findValue = findControls(notepadWindow, wantedClass="Edit")[0]
    _x = input('->')                               
    print("Enter some text - and wait long enough for it to be seen")
    setEditText(findValue, "Hello, mate!")
    time.sleep(.5)
    _x = input('->')
    print("Locate notepad's edit area, and enter various bits of text.")
    editArea = findControl(notepadWindow,wantedClass="Edit")
    setEditText(editArea, "Hello, again!")
    time.sleep(.5)
    setEditText(editArea, "You still there?")
    time.sleep(.5)
    setEditText(editArea, ["Here come", "two lines!"])
    time.sleep(.5)
    _x = input('->')
    
    print("Add some...")
    setEditText(editArea, ["", "And a 3rd one!"], append=True)
    time.sleep(.5)
    
    print("See what's there now:")
    pprint.pprint(getEditText(editArea))
    _x = input('->')
    
    print("Exit notepad")
    activateMenuItem(notepadWindow, ('bestand', 'afsluiten'))  # was 'file', 'exit'
    time.sleep(.5)
    
    print("Don't save.")
    saveDialog = findTopWindow(selectionFunction=lambda hwnd:
                                                 win32gui.GetWindowText(hwnd)=='Notepad')
    noButton = findControl(saveDialog,wantedClass="Button", wantedText="nee")  # was 'no'
    clickButton(noButton)
    _x = input('->')

def test_with_dragonpad1(setText=1):
    dragonpadWindow = findTopWindow(wantedClass='TalkpadClass')
    print('dragonpadWindow: %s'% dragonpadWindow)
    findValues = findControls(dragonpadWindow, wantedClass="RichEdit20A")
    if not findValues:
        return None
    findValue = findValues[0]
    print('findValue: %s'% findValue)
    if setText:
        setEditText(findValue, "Hello, mate!")
    return findValue

def test_with_notepad1(setText=1):
    notepadWindow = findTopWindow(wantedClass='Notepad')
    print('notepadWindow: %s'% notepadWindow)
    if not notepadWindow:
        return None
    findValue = findControls(notepadWindow, wantedClass="Edit")[0]
    print('findValue: %s'% findValue)
    if setText:
        setEditText(findValue, "Hello, mate!")
    return findValue
    # do "switch to notepad" and dictate "this is a test"
    # result:
    # Hello, mate! Hello, this is another test

def test_with_wordpad1():
    wordpadWindow = findTopWindow(wantedClass='WordPadClass')
    print('wordpadWindow: %s'% wordpadWindow)
    findValue = findControls(wordpadWindow, wantedClass="RICHEDIT50W")[0]
    print('findValue: %s'% findValue)
    if not findValue:
        return None
    if setText:
        setEditText(findValue, "Hello, mate!")
    return findValue
    # do "switch to wordpad" and dictate "this is a test"
    # result:
    # Hello, mate! Hello, this is a test

def test_with_excel():
    excelWindows = findTopWindows(wantedClass='XLMAIN')
    for xl in excelWindows:
        print("xl: %s"% xl)
        contents = dumpWindow(xl)
        pprint.pprint(contents)
        
    return excelWindows

def test_with_pythonwin(topHwnd):
    """pass top handle, use selection function from windowsparameter.py
    """
    import windowparameters
    selFunc = windowparameters.getPythonwinEditControl
    controls = findControls(topHwnd, selectionFunction=selFunc)
    print('controls pythonwin: %s'% controls)

    contents = dumpWindow(topHwnd)
    pprint.pprint(contents)
    return controls

def test_with_komodo_messages():
    """also TCP model could be set up, this is the try via scintilla window and messages
    
    discussion on: http://community.activestate.com/node/10359
    
    """
    komodoWindows = findTopWindows(wantedClass='MozillaWindowClass')
    for ko in komodoWindows:
        contents = dumpWindow(ko)
        if contents:
            print("ko: %s"% ko)
            pprint.pprint(contents)
        
    # do "switch to wordpad" and dictate "this is a test"
    # result:
    # Hello, mate! Hello, this is a test

def doHotshotTest(editControl, nTests=6):
    """do some testing on control.
    
    With each higher value of nTests the amount of text is doubled (in the last set/get calls)
    """
    import hotshot, hotshot.stats
    filePath = r'D:\messagesfunctions.prof'
    prof = hotshot.Profile(filePath)
    prof.runcall(hotshot_test, editControl, nTests)
    prof.close()
    stats = hotshot.stats.load(filePath)
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(20)

def doHotshotTestField(editControl):
    """do some testing on control.
    
    With each higher value of nTests the amount of text is doubled (in the last set/get calls)
    """
    import hotshot, hotshot.stats
    filePath = r'D:\messagesfunctions.prof'
    prof = hotshot.Profile(filePath)
    prof.runcall(hotshot_test_selection, editControl, 21, 23)
    prof.close()
    stats = hotshot.stats.load(filePath)
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(20)

def hotshot_test_selection(editArea, selStart, selEnd):
    """do some actions in hotshot profile mode, when a field is selected
    """
    setSelection(editArea, selStart, selEnd)
    time.sleep(1)
    _t = getEditText(editArea)
    time.sleep(1)
    print('selection at: %s, %s'% (selStart, selEnd))

def doHotshotTestOutsideField(editControl):
    """do some testing on control.
    
    With each higher value of nTests the amount of text is doubled (in the last set/get calls)
    """
    import hotshot, hotshot.stats
    filePath = r'D:\messagesfunctionsoutside.prof'
    prof = hotshot.Profile(filePath)
    prof.runcall(hotshot_test_selection, editControl, 10, 20)
    prof.close()
    stats = hotshot.stats.load(filePath)
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(20)

    
def hotshot_test(editArea, nTests):
    """do some actions in hotshot profile mode
    """
    setEditText(editArea, "")
    setEditText(editArea, "Hello, again!")
    setEditText(editArea, "You still there?")
    #setEditText(editArea, ["Here come", "two lines! The second one is not too short!"])
    #setEditText(editArea, ["", "And a 3rd one! This one is considerable longer, because there must be something to test. I hope this testing gives some reliable result.", ""], append=True)
    setEditText(editArea, ["Here come", "two lines! The second one is not too short!"])
    setEditText(editArea, ["", "", "And a 3rd one! This one is considerable longer, because there must be something to test. I hope this testing gives some reliable result.", "", ""], append=True)


    for _i in range(nTests):
        t = getEditText(editArea)
        setEditText(editArea, t, append=True)
    result = getEditText(editArea)
    lines = len(result)
    lenresult = len(''.join(result))
    print('***number of tests: %s, lines: %s, characters: %s  ***'% (nTests, lines, lenresult))

def findPythonwinControl(topHwnd):
    """pass the top handle"""
    windowTitle = win32gui.GetWindowText(topHwnd)
    wantedText = windowTitle.split("[")[-1][:-1]
    result = []
    controls1 = findControls(topHwnd,wantedText=wantedText)
    if controls1:
        result.append(findAdditionalControls(controls1, wantedClass="Scintilla"))
    print('pythonwin: %s'% result)
    


def findDragonBar():
    db = findTopWindow(wantedClass='DgnBarMainWindowCls')
    
def testExplorerViews(hndle):
    # PostMessage, 0x111, 28713,0,, ahk_pid %Win_pid% medium icons, AHK
    # PostMessage = ctypes.windll.user32.PostMessageA
    win32gui.PostMessage(hndle, 28713, 0, 0)
    win32gui.PostMessage(hndle, 28716, 0, 0)
    
        
if __name__ == '__main__':
    #findDragonBar()
    
    #self_test()
    #editArea = test_with_aligen1(setText=0)
    #editArea = test_with_dragonpad1(setText=0)
    #editArea = test_with_notepad1()
    #editArea = test_with_wordpad1()
    #if not editArea:
    #    print('did not find the correct control, ensure your application is on!!'
    #else:
    #    #doHotshotTest(editArea, 6)
    #    doHotshotTestField(editArea)
    #    doHotshotTestOutsideField(editArea)
    
    #getFolderFromCabinetWClass(1180166)
    #test_with_excel()
    #test_with_komodo_messages()
    #findPythonwinControl(133488)  # uses selectionFunction getPythonwinEditControl
    #test_with_pythonwin(133488)
    #pprint.pprint(dumpXplorer2())
    
    # view settings explorer (this one fails!)
    # testExplorerViews(986418)
    # g = getFolderFromCabinetWClass(1837622)
    # g = getFolderFromDialog(394720, '#32770')
    # print(g, type(g))
    # hndle = getForegroundWindow()
    # SendKeys(hndle, "{shift+a}")   ## fails
    # SendKeys(hndle, "abc")         ## works
    pass


#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestMessagefunction.py
# this script tests the message functions, needed for communication with
# a (non active) window.
# Developed by Quintijn Hoogenboom, for Jason Koller, application for global
# dictation caption and putting the results into a fixed target window.
#
# below fill in parameters for finding correct window:
#  tester: Name of the test you are performing...
#  application (testApp)
#  window title (testCaption) (case insensitive, may be part of, or None)
#  className of app (testClass) (case sensitive)
#  W["testcloseapp"] (1: close at each test, 0: leave open after each test)
#      (text after each line, '' in Notepad,
#                                      '\r' in WordPad,
#                                      '\r\n' in aligen ????)
#               (now ignore, getting only the text without \r and \n characters)
#  linesep is important for synchronising the select and say functions.
#                 '\n' Notepad,
#                 '\r\n' for WordPad, aligen (???)

# increase or decrease for more visible or faster testing:
visibleTime = 0.5 # set higher, if you want to see what happens, 1, 2...

import sys
import os
import os.path
import win32clipboard
from pathqh import path

#need this here (hard coded, sorry) for it can be run without NatSpeak being on
extraPaths = [path(r"C:\natlinkGIT3\unimacro")]
for extraPath in extraPaths:
    extraPath.isdir()
    extraNorm = extraPath.normpath()
    if extraNorm not in sys.path:
        sys.path.append(extraNorm)
# little trick to keep testers apart (do not bother)
try:
    import testermod
    tester = testermod.tester
except ImportError:
    tester = "aligen"

import windowparameters

W = windowparameters.PROGS[tester]

import sys
import unittest
import os
import os.path
import time
import traceback        # for printing exceptions
from struct import pack
import natlink
import win32gui
import win32ui
from messagefunctions import *
import messagefunctions as mess
import TestCaseWithHelpers
class TestError(Exception):pass

#---------------------------------------------------------------------------
class UnittestMessagefunctions(TestCaseWithHelpers.TestCaseWithHelpers):
    def setUp(self):
        self.__class__.ctrl = None
        self.__class__.app = None
        for k,v in list(W.items()):
            setattr(self, k, v)

        self.getEditControl()
        if self.ctrl is None:
            raise TestError("could not get edit handle")
        #content = mess.getEditText(self.ctrl)
        #if len(content) > 1:
        #    print 'content of window:\n|%s|'% repr(content)
    
    def tearDown(self):
        time.sleep(visibleTime)
        if W["testcloseapp"]:
            self.clearAppWindow()
        else:
            if self.app and self.ctrl:
                setEditText(self.ctrl, "", classname=self.classname)
            time.sleep(0.2)



    #def assert_selection_through_clipboard_contents(self, expected, text=""):
    #    """tests the contents of the clipboard"""
    #    text = text or "clipboard contents not as expected"
    #    meHndle = win32gui.GetForegroundWindow()
    #    unimacroutils.SetForegroundWindow(self.app)
    #    time.sleep(0.05)
    #    activateMenuItem(self.app, W["commandcut"])
    #    time.sleep(0.05)
    #    t = getClipboard()
    #    unimacroutils.SetForegroundWindow(meHndle)
    #    
    #    text = text + '\nExpected: %s\nGot: %s\n'% (expected, t)
    #    self.assert_(t == expected, text)


    def wait(self, t=None):
        time.sleep(t or 0.1)

    #def getAppWindow(self):
    #    appWindows = findTopWindows(wantedClass=W["windowclass"], wantedText=W["windowcaption"])
    #    if len(appWindows) > 1:
    #        print 'warning, more appWindows active! %s'% appWindows
    #        self.__class__.app = appWindows[0]
    #    elif appWindows:
    #        self.__class__.app = appWindows[0]
    #    else:
    #        if not W["shouldstartauto"]:
    #            raise TestError("application (class: %s, caption: %s) should be active before you start the tests"%
    #                            (W["windowclass"], W["windowcaption"]))
    #        meHndle = win32gui.GetForegroundWindow()
    #        result = os.startfile(W["apppath"])
    #        if W["testcloseapp"]:
    #            sleepTime = 0.1
    #        else:
    #            sleepTime = 0.5
    #        for iTry in range(20):
    #            time.sleep(sleepTime)
    #            appWindows = findTopWindows(wantedClass=W["windowclass"], wantedText=W["windowcaption"])
    #            if appWindows: break
    #        else:
    #            print 'starting %s failed, dumping apps'% testApp 
    #            pprint.pprint(dumpTopWindows())
    #            return
    #
    #        if meHndle:
    #            try:
    #                unimacroutils.SetForegroundWindow(meHndle)
    #            except:
    #                pass
    #        self.__class__.app = appWindows[0]
    def getAppWindow(self):
        appWindow = findTopWindow(wantedClass=W["windowclass"], wantedText=W["windowcaption"])
        if appWindow:
            isVisible = win32gui.IsWindowVisible(appWindow)
            #print 'is visible? %s'% isVisible
            self.__class__.app = appWindow
            return
        if not W["shouldstartauto"]:
            raise TestError("application (class: %s, caption: %s) should be active before you start the tests"%
                            (W["windowclass"], W["windowcaption"]))
        meHndle = win32gui.GetForegroundWindow()
        result = os.startfile(W["apppath"])
        if W["testcloseapp"]:
            sleepTime = 0.1
        else:
            sleepTime = 0.5
        for iTry in range(20):
            time.sleep(sleepTime)
            appWindow = findTopWindow(wantedClass=W["windowclass"], wantedText=W["windowcaption"])
            if appWindow: break
        else:
            print('starting %s failed, dumping apps'% testApp) 
            pprint.pprint(dumpTopWindows())
            return

        if meHndle:
            try:
                unimacroutils.SetForegroundWindow(meHndle)
            except:
                pass
        self.__class__.app = appWindow
            
    def getEditControl(self):
        # get application, if not found, set ctrl to None
        self.getAppWindow()
        if not self.app:
            self.__class__.ctrl = None
            return
        if tester == 'pythonwin':
            wTitle = win32gui.GetWindowText(self.app)
            filename = wTitle.split('[')[-1][:-1]  # remove [ and ], leaving only the filenam
            def selectionFunction(hndle, gotTitle, gotClass):
                """special for selecting the Afx class with the same title as the complete window title bar
                being the filename in question
                
                special for pythonwin and only for the FIRST search action for child windows.
                After the correct Afx window has been identified, the Scintilla child window is the correct one.
                """
                if gotTitle == filename:
                    #print 'got afx with title: %s'% gotTitle
                    return True
        else:
            selectionFunction = None
        currentSel = None
        if self.ctrl:
            currentSel = mess.getSelection(self.ctrl)
        if currentSel and currentSel != (0,0):
            return
        wC, wT = W["editcontrol"], W["edittext"]
        choiceControl = 0
        if type(wT) in [type(None), bytes, int]: wT =  [wT]
        if type(wC) in [type(None), bytes, int]: wC =  [wC]
        ctrl = self.app            
        for wtext, wclass in zip(wT, wC):
            ctrls =  mess.findControls(ctrl,wantedText=wtext, wantedClass=wclass,selectionFunction=selectionFunction)
            if selectionFunction:
                selectionFunction = None # execute only for first findControls action pythonwin
                if tester == 'pythonwin':
                    choiceControl = -1

            if len(ctrls):
                ctrl = ctrls[choiceControl]
                #print 'editHndle set to: %s'% editHndle
                if len(ctrls) > 1:
                    for hndle in ctrls:
                        id = win32gui.GetDlgCtrlID(hndle)
                        if id == 0xd6:
                            ctrl = hndle
                            break
            else:
                pprint.pprint(mess.dumpWindow(self.app))
                raise ValueError("could not find the editHndle of the control: %s in application: %s"%
                              (self.editcontrol, self.apppath))
                
        self.__class__.ctrl = ctrl
        self.__class__.classname = win32gui.GetClassName(ctrl)

    def clearAppWindow(self):
        if self.app:
            time.sleep(visibleTime)
            setEditText(self.ctrl, '', classname=self.classname)
            quitProgram(self.app)
            self.app = None
            time.sleep(0.1)

    def dumpWindows(self):
        """dump non-standard window names"""
        dumpTopWindows(all=None)
    #---------------------------------------------------------------------------
    def test_get_controls_of_application(self):
        self.assertTrue(self.app > 0, "app should be there now")
        if self.ctrl:
            print('edit area: %s'% self.ctrl)
        else:
            print('all controls of app %s------'% testApp)
            print('all controls of application:-----------------')
            pprint.pprint(dumpWindow(self.app))
        
    def test_existence_of_application(self):
        self.assertTrue(self.app > 0, "app (handle to applicatione) should be there now")
        #print 'all top windows:------------'
        #pprint.pprint(dumpTopWindows())
        # try closing and then through exception again:
        # only if you choose to close app after each test:
        # this raises an error in tearDown, because the app is already closed then...
        # maybe sort out later.
        #if W["testcloseapp"]:
        #    quitProgram(self.app)
        #pass
        self.assertTrue(self.ctrl > 0, "ctrl (handle to edit control) should be there now")
        print('content of window at start:\n|%s|'% mess.getEditText(self.ctrl))
        
    def test_set_text_in_application(self):
        """setting and getting text in application
        
        setting can be done as a string or as a list of strings.
        In the latter case newlines are inserted.
        """
        sep = self.linesep
        self.assertTrue(self.app > 0, "app should be there now")
        gotText = getEditText(self.ctrl)
        self.assert_equal([sep], gotText, 'text of edit box should be empty at start, adjust W["linesep"] or W["aftertext"] in windowparameters!')

        # setting "hello there"
        setText = "hello there"
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        expected = [setText+sep]
        # so aftertext is pasted after the text!
        self.assert_equal(expected, gotText, "text of edit box should be the same as the set text")

        # settting "in Amsterdam"
        setText = "in Amsterdam"
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        expected = [setText+sep]
        self.assert_equal(expected, gotText, "text of edit box should be the same as the set text")
        
        # appending ", hello there."
        setText = ", hello there."
        setEditText(self.ctrl, setText,append=True)
        gotText = getEditText(self.ctrl)
        expected = ["in Amsterdam, hello there."+sep]
        self.assert_equal(expected, gotText, "text of edit box should be the same as the set text")
        
        # appending a new line
        setText = ["", "", "How are you doing?", ""]
        setEditText(self.ctrl, setText,append=True)
        gotText = getEditText(self.ctrl)
        expected = ['in Amsterdam, hello there.'+sep, sep,
                    'How are you doing?'+sep, sep]
        self.assert_equal(expected, gotText, "text of edit box should be the same as the set text")

        # clear all and append as one string with \r
        self.clearAppWindow()
        sep2 = sep + sep
        setText = sep2.join(["first", "third line"])
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        expected = ["first"+sep, sep, "third line"+sep]
        self.assert_equal(expected, gotText, "text of edit box should be the same as the set text (sent as one string")
        
    def test_selection_set_text_in_application(self):
        """setting and getting text in application in a selected range
        
        """
        sep = self.linesep
        self.assertTrue(self.app > 0, "app should be there now")
        # setting "hello there"
        setText = "hello there"
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        expected = [setText+sep]
        self.assert_equal(expected, gotText, "text of edit box should contain initial text now")
        # get selection (at end)
        
        # set selection:
        lo, hi = getSelection(self.ctrl)
        expected = len(setText), len(setText)
        self.assert_equal( expected, (lo,hi), "selection returned differs from trying to set")
        
        setSelection(self.ctrl, 1,4)  # "ell" 
        lo, hi = getSelection(self.ctrl)
        self.assert_equal( (1,4), (lo,hi), "selection returned differs from trying to set")

        # append a second paragraph:
        setText = ["ELLLLLL"]
        replaceEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        
        expected = ['hELLLLLLo there'+sep]
        self.assert_equal(expected, gotText, "text of edit box after insert in selection is wrong")

        lo, hi = getSelection(self.ctrl)
        expectedRange = (8,8)  #  1 + 7,  start of previous + length of insertion
        self.assert_equal(expectedRange, (lo, hi), "selection after insertion not as expected")
        #clearClipboard()
        activateMenuItem(self.app, W["commandselectall"])
        lo, hi = getSelection(self.ctrl)
        expectedSel = (0, len(expected[0]))
        self.assert_equal(expectedSel, (lo, hi), "selection after select all not as expected")


    def test_scrolling_text_in_application(self):
        """see what happens if text scrolls in a control"""
        sep = self.linesep
        self.assertTrue(self.app > 0, "app should be there now")
        # setting text:
        longerText = "This text is longer than 80 characters, so should scroll in most controls. I therefore like to see what happens with this text."
        secondPara = "This goes on the next line. Is also considerable of length, so we can find the differences between scrolled lines and real new lines."
        setEditText(self.ctrl, [longerText, secondPara] )
        gotText = getEditText(self.ctrl)
        self.assertTrue(len(gotText) > 1, "text should scroll, make window more narrow")
        
        expected = ''.join([longerText+sep, secondPara+sep])
        got = ''.join(gotText)
        self.assert_equal(expected, got, "text of edit box should contain initial text now")
        # get selection (at end)

    def test_getting_more_info_from_application(self):
        """try to get text length and other things
        
        this is quite a messy test, used also for timing purposes.
        Need a U: (virtual) drive for testing (with hotspot) and use the small print_hotspot program to
        print the results
        """
        sep = self.linesep
        aft = self.aftertext
        self.assertTrue(self.app > 0, "app should be there now")
        # setting text:
        longerText = "This text is longer than 80 characters, so should scroll in most controls. I therefore like to see what happens with this text. I hope to see the scrolling taking place, without updating the line number. I type on and on in the hope that even a very wide control cannot handle this amount of text in one line."
                     
        total = []
        for i in range(100):
            total.append("This is a sligthly longer line %s."% (i+1,))
        total.append(longerText)
        setEditText(self.ctrl, sep.join(total) )
        gotText = getEditText(self.ctrl)
        self.assertTrue(len(gotText) > 1, "text should scroll, make window more narrow")
        
        expected = sep.join(total) + aft
        got = ''.join(gotText)
        self.assert_equal(expected, got, "text of edit box should contain initial text now")
        
        numLines = getNumberOfLines(self.ctrl)
        self.assert_equal(101, numLines, "number of lines in control does not match")
        # line number of call is one less than on the screen (0 based, while screen is 1 based)
        line = getTextLine(self.ctrl, 1)
        expected = 'This is a sligthly longer line 2.\r'
        self.assert_equal(expected, line, "text of line 2 (internal 1) does not match expected")

        line = getTextLine(self.ctrl, 20)
        expected = 'This is a sligthly longer line 21.\r'
        self.assert_equal(expected, line, "text of line 21 (internal 20) does not match expected")


        line = getTextLine(self.ctrl, 0)
        expected = 'This is a sligthly longer line 1.\r'
        self.assert_equal(expected, line, "text of line 1 (internal 0) does not match expected")


        line = getTextLine(self.ctrl, 123456)
        expected = ''
        self.assert_equal(expected, line, "text of line VERY_LARGE_NUMBER does not match expected")

        lastLine = getTextLine(self.ctrl, numLines-1)
        expected = 'like to see what happens with this text.\r'
        self.assert_equal(expected, lastLine, "text of last line does not match expected")

        setSelection(self.ctrl, 10, 50)
        ln = getLineNumber(self.ctrl) # of cursor position
        setSelection(self.ctrl, 500, 500)
        ln = getLineNumber(self.ctrl)
        setSelection(self.ctrl, 5000, 5001)
        ln = getLineNumber(self.ctrl)
        setSelection(self.ctrl, 5000, 4000)
        ln = getLineNumber(self.ctrl)
        setSelection(self.ctrl, 5000, 400000)
        ln = getLineNumber(self.ctrl)


        #getStartEndOfLine(self.ctrl, ln)
        info = getNumberOfLines(self.ctrl)
        print('info: %s'% repr(info))
        for i in range(1):
            rawLength = getRawTextLength(self.ctrl)
            gotTextAll, bufLen = getWindowTextAll(self.ctrl, rawLength=rawLength)
        print('rawLength: %s'% rawLength)
        # get selection (at end)
        gotText = ''.join(getEditText(self.ctrl))
        print('length of buffer: %s'% len(gotText))
        #print 'buffer (old way): %s'% repr(gotText)


        gotTextAll, bufLen = getWindowTextAll(self.ctrl)
        setEditText(self.ctrl, gotTextAll)
        gotTextAll2, bufLen2 = getWindowTextAll(self.ctrl)
        print('second All== firstAll? %s'% (gotTextAll == gotTextAll2))
        print('gotText (%s): %s'% (len(gotText), repr(gotText[:100])))
        print('gotText2 (%s): %s'% (len(gotTextAll2), repr(gotTextAll2[:100])))
        setEditText(self.ctrl, gotTextAll2)
        gotText2 = ''.join(getEditText(self.ctrl))
        setEditText(self.ctrl, gotText2)
        gotText3 = ''.join(getEditText(self.ctrl))
        print('gotText2 (%s): %s'% (len(gotTextAll2), repr(gotTextAll2[:100])))
        print('Org: 1 == 2? %s'% (gotText == gotText2))
        print('Org: 2 == 3? %s'% (gotText3 == gotText2))

        #for i in range(5000):
        #    total.append("This is another still sligthly longer line %s."% (i+1,))
        #setEditText(self.ctrl, sep.join(total) )
        #gotTextAll, bufLen = getWindowTextAll(self.ctrl)
        #print 'buflen after more lines: %s'% bufLen
        
        
    def test_selection_set_text_in_application(self):
        """setting and getting text in application in a selected range
        
        """
        sep = self.linesep
        self.assertTrue(self.app > 0, "app should be there now")
        # setting "hello there"
        setText = "hello there"
        setEditText(self.ctrl, setText, classname=self.classname)
        gotText = getEditText(self.ctrl, self.classname)
        expected = [setText+sep]
        self.assert_equal(expected, gotText, "text of edit box should contain initial text now")
        
        # get selection (at end)
        
        # set selection:
        lo, hi = getSelection(self.ctrl, classname=self.classname)
        expected = len(setText), len(setText)
        self.assert_equal( expected, (lo,hi), "selection returned differs from trying to set")
        
        setSelection(self.ctrl, 1,4, classname=self.classname)  # "ell" 
        lo, hi = getSelection(self.ctrl)
        self.assert_equal( (1,4), (lo,hi), "selection returned differs from trying to set")

        # append a second paragraph:
        setText = ["ELLLLLL"]
        replaceEditText(self.ctrl, setText, classname=self.classname)
        gotText = getEditText(self.ctrl)
        
        expected = ['hELLLLLLo there'+sep]
        self.assert_equal(expected, gotText, "text of edit box after insert in selection is wrong")

        lo, hi = getSelection(self.ctrl, classname=self.classname)
        expected = (8,8)  #  1 + 7,  start of previous + length of insertion
      
        self.assert_equal(expected, (lo, hi), "selection after insertion not as expected")

        setText = [" at the bottom"]
        appendEditText(self.ctrl, setText, classname=self.classname)
        gotText = getEditText(self.ctrl)
        
        expected = ['hELLLLLLo there at the bottom'+sep]
        self.assert_equal(expected, gotText, "text of edit box after insert in selection is wrong")

        lo, hi = getSelection(self.ctrl, classname=self.classname)
        expected = (29, 29)  
        self.assert_equal(expected, (lo, hi), "selection after insertion not as expected")

        setText = [" longer text, because there should be a scroll now, depending on the settings of wordpad, notepad or aligen."]
        appendEditText(self.ctrl, setText, classname=self.classname)
        gotText = getEditText(self.ctrl)
        expected = ['hELLLLLLo there at the bottom'+sep]
        lo, hi = getSelection(self.ctrl, classname=self.classname)
        expectedSelection = (137, 137)  
        self.assert_equal(expectedSelection, (lo, hi), "selection after insertion not as expected")
        expected = ['hELLLLLLo there at the bottom longer ', 'text, because there should be a scroll ', 'now, depending on the settings of ', 'wordpad, notepad or aligen.']
        self.assert_equal(''.join(expected)+sep, ''.join(gotText), "longer text of edit box after insert in selection is wrong (Normal for notepad in different sizes)")


    def test_visible_line_in_application(self):
        """getting the first visible line in an application
        
        """
        sep = self.linesep
        
        self.assertTrue(self.app > 0, "app should be there now")
        # setting "hello there"
        L = []
        for i in range(100):
            L.append(str(i) + " " + "A"*50 + " " + "B"*50)
            
        setEditText(self.ctrl, L)
        lineNo = getFirstVisibleLine(self.ctrl)
        expectedMin = 100
        self.assertTrue(lineNo > expectedMin, "lineNo should be at least %s, is now: %s"%
                     (expectedMin, lineNo))
        print('first visible line in app: %s'% lineNo)
        gotText = getEditText(self.ctrl, visible=True)
        pass

    def test_menu_functions(self):
        """test select all, copy and cut functions
        
        """
        sep = self.linesep
        editcontrol = self.editcontrol
        
        self.assertTrue(self.app > 0, "app should be there now")
        # first selectall with empty window:
        gotText = getEditText(self.ctrl)
        expected = [sep]
        self.assert_equal(expected, gotText, "empty edit box should contain only aftertext now")
        time.sleep(0.1)
        lo, hi = getSelection(self.ctrl)
        expSel = (0, 0)  # or (0, len(sep)) now it seems the selection is WITHOUT the trailing \r or \r\n characters
        self.assertTrue(expSel == (lo,hi), "selection initially empty screen not as expected\nExpected: %s\nActual: %s"%
                     (repr(expSel), repr((lo, hi))))
        #clearClipboard()
        activateMenuItem(self.app, W["commandselectall"])
        time.sleep(0.1)
        lo, hi = getSelection(self.ctrl)
        if sep == '\r\n':
            expSel = (0,  len(sep))  # aligen
        else:                        # differs from
            expSel = (0, 0)          # wordpad
        self.assertTrue(expSel == (lo,hi), "selection empty screen after command Select All not as expected\nExpected: %s\nActual: %s"%
                     (repr(expSel), repr((lo, hi))))
        
        #now put some text in and check:
        setText = "text for selection"
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        expected = [setText+sep]
        self.assert_equal(expected, gotText, "text of edit box should contain initial text now")
        #clearClipboard()
        activateMenuItem(self.app, W["commandselectall"])
        time.sleep(0.1)
        lo, hi = getSelection(self.ctrl)
        # for one line the selection is WITHOUT the trailing \r character!! (wordpad)
        if sep == '\r\n':
            expSel = (0,len(setText+sep))   # again aligen
        elif editcontrol == "RichEdit20A":
            # DragonPad, new aligen
            expSel = (0,len(setText+sep))     # DragonPad  (sep is \r)
        else:
            expSel = (0,len(setText))       # wordpad
        self.assertTrue(expSel == (lo,hi), "selection after select all not as expected, check W['commandselectall']\nExpected: %s\nActual: %s"%
                     (repr(expSel), repr((lo, hi))))

        setText = ["", "more", "", "", "lines", "", "", "of", "", "", "text.", ""]
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        print('gotText: %s'% repr(gotText))
        expected = [s+sep for s in setText]
        self.assert_equal(expected, gotText, "text of edit box should contain more lines of text now")
        #clearClipboard()
        activateMenuItem(self.app, W["commandselectall"])
        time.sleep(0.1)
        lo, hi = getSelection(self.ctrl)
        hiExp= sum(len(s) for s in expected) 
        expSel = (0,hiExp)
        self.assertTrue(expSel == (lo,hi), "selection after select all not as expected (trailing empty line), check W['commandselectall']\nExpected: %s\nActual: %s"%
                     (repr(expSel), repr((lo, hi))))

        setText = ["", "more", "", "", "lines"]
        setEditText(self.ctrl, setText)
        gotText = getEditText(self.ctrl)
        expected = [s+sep for s in setText]
        self.assert_equal(expected, gotText, "text of edit box (no trailing empty line) should contain initial text now")
        #clearClipboard()
        activateMenuItem(self.app, W["commandselectall"])
        time.sleep(0.1)
        lo, hi = getSelection(self.ctrl)
        hiExp= sum(len(s) for s in expected) 
        # now the selection is INCLUDING the trailing \r character!!
        expSel = (0,hiExp)
        self.assertTrue(expSel == (lo,hi), "selection after select all not as expected (no trailing em, check W['commandselectall']\nExpected: %s\nActual: %s"%
                     (repr(expSel), repr((lo, hi))))

        


    def test_sending_keystrokes(self):
        sep = self.linesep
        #sendKey(self.ctrl, sep) # fails for aligen (\r\n)
        sendKey(self.ctrl, "aA")
        time.sleep(0.1)
        sendKey(self.ctrl, "{left}")
        time.sleep(0.1)
        sendKey(self.ctrl, "zZ")
        time.sleep(0.1)
        gotText = getEditText(self.ctrl)
        expected = ["azZA" + sep]
        self.assertTrue(expected == gotText, "sendking keystrokes give unexpected tesult\nExpected: %s\nActual: %s"%
                     (repr(expected), repr(gotText)))

        # trying delete key:
        sendKey(self.ctrl, "{delete}")
        gotText = getEditText(self.ctrl)
        expected = ["azZ" + sep]
        self.assertTrue(expected == gotText, "sendking keystrokes (DELETE) give unexpected tesult\nExpected: %s\nActual: %s"%
                     (repr(expected), repr(gotText)))
        # trying backspace (back) key:
        sendKey(self.ctrl, "{back}")
        gotText = getEditText(self.ctrl)
        expected = ["az" + sep]
        self.assertTrue(expected == gotText, "sendking keystrokes (BACK, backspace) give unexpected tesult\nExpected: %s\nActual: %s"%
                     (repr(expected), repr(gotText)))

        # extra little trick:
        sendKey(self.ctrl, "{backspace}")
        gotText = getEditText(self.ctrl)
        expected = ["a" + sep]
        self.assertTrue(expected == gotText, "sendking keystrokes (BACK, backspace) give unexpected tesult\nExpected: %s\nActual: %s"%
                     (repr(expected), repr(gotText)))

        # now try to delete selection:
        sendKey(self.ctrl, "abcdefgh")
        time.sleep(0.1)
        activateMenuItem(self.app, W["commandselectall"])
        time.sleep(0.1)
        sendKey(self.ctrl, '{del}')
        time.sleep(0.1)
        gotText = getEditText(self.ctrl)
        expected = [sep]
        self.assertTrue(expected == gotText, "sendking keystrokes (delete) should have emptied the window\nExpected: %s\nActual: %s"%
                     (repr(expected), repr(gotText)))

def try_win32pad():
    """needs win32pad being open and with a file in, explores the line number facilities
    """
    appWindows = findTopWindows(wantedClass="win32padClass")
    print('# appWindows: %s'% len(appWindows))

    if len(appWindows) != 1:
        print('no win32pad found, or more instances: %s'% len(appWindows))
        return
    
    appWindow = appWindows[0]
    title = win32gui.GetWindowText(appWindow)
    print('appWindow win32pad: %s (%s)'% (appWindow, title))
    pprint.pprint(dumpWindow(appWindow))


    
           
def try_32770():
    """dialog for open etc from office etc
    
    
    only print details if Adres: of Address: is found...
    """
    appWindows = findTopWindows(wantedClass="#32770")
    print('# appWindows: %s'% len(appWindows))
    for appWindow in appWindows:
        title = win32gui.GetWindowText(appWindow)
        print('appWindow 32770: %s (%s)'% (appWindow, title))
    #    
    #    
    #for appWindow in appWindows:
    ##testControl = 723600
    ##if not testControl in appWindows:
    ##    print 'testup new test for your purposes'
    ##    return
    #pprint.pprint(dumpWindow(testControl))
    
        controls = findControls(appWindow, selectionFunction=selFuncExplorerAddress)
        if controls:
            print('controls: %s'% controls)
            hndle = controls[0]
            text = win32gui.GetWindowText(hndle)
            folder = text.split(": ", 1)[1]
            if os.path.isdir(folder):
                print('ok: %s'% folder)
            else:
                print('no folder: %s'% folder)
            pprint.pprint(dumpWindow(appWindow))
        #else:
    pprint.pprint(dumpWindow(appWindow))
            

    
    #controls = findControls(appWindow, selectionFunction=selFuncExplorerAddress)
    #if controls:
    #    hndle = controls[0]
    #    text = win32gui.GetWindowText(hndle)
    #    folder = text.split(": ", 1)[1]
    #    print folder
    #    if os.path.isdir(folder):
    #        print 'ok: %s'% folder
    #


        
        
def run_hotshot():
    import hotshot, hotshot.stats
    filePath = r'U:\messagesfunctionstest.prof'
    prof = hotshot.Profile(filePath)
    prof.runcall(run)
    prof.close()
    # printing with print_hotshot in miscqh...       
        

def run():
    print('starting unittestMessagefunctions')
    unittest.main()
    

if __name__ == "__main__":
    run()
    #run_hotshot()
    #try_explorer()  #open explorer window and run
    #try_32770()  # probeert alle 32770 windows en geeft van een open dialog window het interne pad
    
    #try_win32pad()


"""actions classes for specific programs, base classes first

This module provides the different classes to base a program specific actionsclass on

"""
#pylint:disable=C0116, W0201
from dtactions import messagefunctions as mess
from dtactions.unimacro import unimacroutils

class AllActions:
    """base class for all actions that run in actionclasses
    """
    def __init__(self, progInfo):
        self.reset(progInfo=progInfo)
        
    # init same as reset:        
    def reset(self, progInfo=None):
        if progInfo is None:
            progInfo = unimacroutils.getProgInfo()
        self.progInfo = progInfo
        
    def update(self, newProgInfo=None):
        if newProgInfo is None:
            newProgInfo = unimacroutils.getProgInfo()
        if newProgInfo == self.progInfo:
            return
        # print('allactions: new prog info, overload for your specific program: %s'% self.prog)
        self.progInfo = newProgInfo
        
    def getCurrentLineNumber(self):
        pass
    
# base actions for applications that are connected through Windows Messages Functions
class MessageActions(AllActions):
    """class based upon windows messages, to get information about its status
    """
    def __init__(self, progInfo):
        AllActions.__init__(self, progInfo=progInfo)
        self.classname = None   #  Scintilla can be handled quite generically with setting this variable
        self.prevTabName = None
        self.update(progInfo)

    def getInnerHandle(self, topHndle):
        """should be overridden by a specific Messages window class
        """
        raise NotImplementedError

    def update(self, newProgInfo):
        """if prog info changes, then probably only title changes so notice and do nothing
        """
        if newProgInfo == self.progInfo:
            return 1 # OK
        if not newProgInfo:
            print('actionabases, update, no new ProgInfo')
            return 0
        self.progInfo = newProgInfo
        self.ctrl = self.handle = self.getInnerHandle(self.progInfo.hndle)
        if not self.handle:
            if newProgInfo and newProgInfo.toporchild == 'top':
                print('no handle found for (top) edit control for program: %s'% self.progInfo.prog)
            return 0
        print(f'updated program info: {self.progInfo}, edit control (inner) handle: {self.handle}')
        return self.handle  # None if no valid handle        

    def getCurrentLineNumber(self, handle=None):
        handle = handle or self.handle
        if not handle:
            return None
        linenumber = mess.getLineNumber(handle, classname=self.classname)  # charpos = -1
        return linenumber

    def getNumberOfLines(self, handle=None):
        nl = mess.getNumberOfLines(handle, classname=self.classname)  
        return nl
    
    def isVisible(self, handle=None):
        handle = handle or self.handle
        if not handle:
            return None
        return mess.isVisible(handle, classname=self.classname)
    
    def getEditText(self, handle=None):
        handle = handle or self.handle
        if not handle:
            return None
        tList = mess.getEditText(self.ctrl)
        return ''.join(tList)
    
    getWindowText = getEditText
        
    def replaceSelection(self, output, handle=None):
        """overwrite selection with output
        """
        handle = handle or self.handle
        if not handle:
            return
        mess.replaceEditText(self.ctrl, output)

    def clearBoth(self, handle=None):
        """clear the target window, and subsequently the dictobj
        """
        handle = handle or self.handle
        if not handle:
            return
        mess.setEditText(handle, "")
        self.updateState()
        
    def getSelection(self, handle=None):
        """get the selection of the edit control
        
        return a 2 tuple
        """
        handle = handle or self.handle
        if not handle:
            return None
        return mess.getSelection(handle)

    def setSelection(self, start, end, handle=None):
        """change the selection of the edit control
        """
        handle = handle or self.handle
        if not handle:
            return
        mess.setSelection(handle, start, end)
    
    def getVisibleRegion(self, handle=None):
        """Utility subroutine which calculates the visible region of the edit

         control and returns the start and end of the current visible region.
         Never got this working
        """
        handle = handle or self.handle
        if not handle:
            return None
        return None, None
        
        # TODOQH    
        # ## to be investigated
        # buf = self.getWindowText()
        # return 0, len(buf)
        
        # top,bottom,left,right = handle.GetClientRect()
        # firstLine = handle.GetFirstVisibleLine()
        # visStart = handle.LineIndex(firstLine)
        # 
        # lineCount = handle.GetLineCount()
        # lastLine = lineCount
        # for line in range(firstLine+1,lineCount):
        #     charInLine = handle.LineIndex(line)
        #     left,top = handle.GetCharPos(charInLine)
        #     if top >= bottom:
        #         break
        #     lastLine = line
        # 
        # visEnd = handle.LineIndex(lastLine+1)
        # if visEnd == -1:
        #     visEnd = len(handle.GetWindowText())
        # return visStart,visEnd
    
    
        


"""utility functions from Quintijn, used in unimacro and in local programs.
   in python3 also the path module as subclass of the standard path class
"""
#pylint:disable=C0116, C0302,  W0613, R0912, R0914, R0915, R0911, W0702
import sys
import unicodedata
import os
import re
import traceback
import collections
import time
from pathlib import Path

# skip the string.maketrans business, ook de fixwordquotes
# _allchars = string.maketrans('', '')
# ## string translator functions:
# def translator(frm=b'', to=b'', delete=b'', keep=None):
#     """closure function to implement the string.translate functie
#     python cookbook (2), ch 1 recipe 9
# obsolete, test for doctest unicode functions...
# 
# # functions below, see also testUtilsqh.py in unittest.
# >>> fixwordquotes(b'\x91aa\x92')
# b"'aa'"
# >>> fixwordquotes(b'\x93bb\x94')
# b'"bb"'
# 
# ## do these via unicode:
# >>> normalizeaccentedchars('d\\u00e9sir\\u00e9 //\u00ddf..# -..e.')
# 'desire //Yf..# -..e.'
# 
# # this one should go before normalizeaccentedchars
# #(and after splitting of the extension and folder parts)
# >>> fixdotslash('abc/-.def this is no extension.')
# 'abc_-_def this is no extension_'
# 
# ## do via unicode:
# ## normalise a inivars key (or section)
# >>> fixinivarskey('abcd')
# 'abcd'
# >>> fixinivarskey("abcd e'f  g")
# 'abcd e_f g'
# >>> fixinivarskey("##$$abcd)e'f  g*")
# 'abcd_e_f g'
# """
#     if len(to) == 1:
#         to = to * len(frm)
#     trans = string.maketrans(frm, to)
#     if keep is not None:
#         delete = _allchars.translate(_allchars, keep.translate(_allchars, delete))
#     def translate(s):
#         return s.translate(trans)
#     return translate
# 
# fixwordquotes = translator(b'\x91\x92\x93\x94\x95\x96', b"''\"\"  ")

###removenoncharacters = translator('
# def fixinivarskey(s):
#     """remove all non letters to underscore, remove leading underscore
#     remove double spaces
#     """
#     if isinstance(s, str):
#         s = str(s)
#     t = translate_non_alphanumerics(s)
#     t = t.strip("_ ")
#     while '  ' in t:
#         t = t.replace('  ', ' ')
#     return t
    
def unifyaccentedchars(to_translate):
    """change acuted characters with combining code to single characters
    
The combining variant (s3) is converted into the one character shorter variant s1, apart from the capitals.
This is done by the NFC variant of the normalize function
    
>>> s1 = "cafcomb\u00e9"   # single char e acute no change
>>> s2 = unifyaccentedchars(s1)
>>> s2
'cafcomb\u00e9'

## this one changes to s1:
>>> s3 = "CafCombe\u0301"   # combining char e acute   0301 is the combining code
>>> s4 = unifyaccentedchars(s3)   # combining char e acute
>>> s4
'CafComb\u00e9'
>>> len(s1), len(s2), len(s3), len(s4)
(8, 8, 9, 8)
>>> s1 == s3.lower()
False
>>> s2 == s4.lower()
True

    (from Fluent Python)
    """
    norm_txt = unicodedata.normalize('NFC', to_translate)
    return norm_txt

def normalizeaccentedchars(to_translate):
    """change acutechars to ascii 
    
>>> s1 = "cafnon\u00e9"   # single char e acute
>>> normalizeaccentedchars(s1)  
'cafnone'
>>> s2 = "CafCombe\u0301"   # combining char e acute   0301 is the combining code
>>> normalizeaccentedchars(s2)   # combining char e acute
'CafCombe'
>>> len(s1), len(s2)
(7, 9)

    (from Fluent Python)
    """
    norm_txt = unicodedata.normalize('NFD', to_translate)
    shaved = ''.join(c for c in norm_txt if not unicodedata.combining(c))
    return shaved

def doubleaccentedchars(to_translate):
    """change acutechars to ascii, but double them e acute = ee
    (from Fluent Python, adaptation QH)
    

>>> s1 = "double caf\\u00e9"   # single char e acute
>>> doubleaccentedchars(s1)
'double cafee'
>>> s2 = "Double Cafe\\u0301"   # combining char e acute
>>> doubleaccentedchars(s2)
'Double Cafee'

>>> doubleaccentedchars("enqu\N{LATIN SMALL LETTER E}\N{COMBINING CIRCUMFLEX ACCENT}te")
'enqueete'

    """
    norm_txt = unicodedata.normalize('NFD', to_translate) ## haal char en accent uit elkaar
    shaved = []
    last = ""
    for c in norm_txt:
        comb = unicodedata.combining(c)
        if comb:
            if comb == 230:
                # accent aegu, accent grave, accent circonflex, decide in favour of accent aegu, double char
                # print('doubleaccentedchars, combining value %s, double char: %s (%s)'% (comb, last, to_translate))
                shaved.append(last)
            elif comb == 202:
                if last == "c":
                    # print('c cedilla, change to "s" (%s)'% to_translate)
                    shaved.pop()
                    shaved.append("s")
                else:
                    # print("c cedilla, but NO C, ignore (%s)"% to_translate)
                    pass
            else:
                print('(yet) unknown combining char %s in "%s", ignore'% (comb, to_translate))
            last = ""
        else:
            shaved.append(c)
            last = c
    return ''.join(shaved)
    # shaved = ''.join(c for c in norm_txt if not unicodedata.combining(c))
    # return shaved
    
# def convertToBinary(unicodeString, encoding=None):
#     """convert a str (unicodeString) to bytes
#     
#     encode encoding (list of strings or string).
#     when encoding is None: take ['ascii', 'cp1252', 'latin-1']
#     
# ## \u0041 is A
# ##unichr(233) or \u00e9 is e accent acute
#     
# # >>> t = '\u0041-xyz-' + unichr(233) + '-abc-'
# >>> t = '\u0041-xyz-\u00e9-abc-'
# >>> convertToBinary(t)
# b'A-xyz-\\xe9-abc-'
# >>> convertToBinary(t+'ascii', 'ascii')
# convertToBinary, cannot convert to printable string with encoding: ['ascii']
# return with "?": b'A-xyz-?-abc-ascii'
# b'A-xyz-?-abc-ascii'
# >>> convertToBinary(t+'cp1252', 'cp1252')
# b'A-xyz-\\xe9-abc-cp1252'
# >>> byteslatin1 = convertToBinary(t+'latin-1', 'latin-1')
# >>> byteslatin1
# b'A-xyz-\\xe9-abc-latin-1'
# >>> bytesutf8 = convertToBinary(t+'utf-8', 'utf-8')
# >>> bytesutf8
# b'A-xyz-\\xc3\\xa9-abc-utf-8'
# >>> convertToBinary(t+'ascii + cp1252', ['ascii', 'cp1252'])
# b'A-xyz-\\xe9-abc-ascii + cp1252'
# >>> convertToBinary(convertToBinary(t+'double convert'))
# b'A-xyz-\\xe9-abc-double convert'
# >>> convertToBinary(byteslatin1)
# b'A-xyz-\\xe9-abc-latin-1'
# >>> convertToBinary(bytesutf8)
# b'A-xyz-\\xe9-abc-utf-8'
# 
# ## \x92 (PU2) is from cp1252 (windows convention): 
# >>> convertToBinary('fondationnimba rapportsd\x92archive index.html')
# b'fondationnimba rapportsd\\x92archive index.html'
#     """
#     # a binary string can hold accented characters:
#     if type(unicodeString) == bytes:
#         unicodeString = convertToUnicode(unicodeString)
#     if encoding is None:
#         encoding = ['ascii', 'cp1252', 'latin-1']
#     elif encoding and type(encoding) in (str, bytes):
#         encoding = [encoding]
#     res = ''
#     for enc in encoding:
#         try:
#             res = unicodeString.encode(enc)
#             break
#         except UnicodeEncodeError:
#             pass
#     else:
#         res = unicodeString.encode('ascii', 'replace')
#         print('convertToBinary, cannot convert to printable string with encoding: %s\nreturn with "?": %s'% (encoding, res))
#     return res

class peek_ahead:
    """ An iterator that supports a peek operation.
    
    Improved for python3 after QH's example: Adapted for python 3 by Paulo Roma.
    
    this is a merge of example 19.18 of python cookbook part 2, peek ahead more steps
    and the simpler example 16.7, which peeks ahead one step and stores it in
    the self.preview variable.
    
    Adapted so the peek function never raises an error, but gives the
    self.sentinel value in order to identify the exhaustion of the iter object.
    
    Example usage (Paulo):
    
    >>> p = peek_ahead(range(4))
    >>> p.peek()
    0
    >>> p.next(1)
    [0]
    >>> p.isFirst()
    True
    >>> p.preview
    1
    >>> p.isFirst()
    True
    >>> p.peek(3)
    [1, 2, 3]
    >>> p.next(2)
    [1, 2]
    >>> p.peek(2) #doctest: +ELLIPSIS
    [3, <object object at ...>]
    >>> p.peek(1)
    [3]
    >>> p.next(2)
    Traceback (most recent call last):
    StopIteration
    >>> p.next()
    3
    >>> p.isLast()
    True
    >>> p.next()
    Traceback (most recent call last):
    StopIteration
    >>> p.next(0)
    []
    >>> p.peek()  #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.preview #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.isLast()  # after the iter process p.isLast remains True
    True
    
    ### my old unittests, QH:
    
        >>> p = peek_ahead(range(4))
    >>> p.peek()
    0
    >>> p.next(1)
    [0]
    >>> p.isFirst()
    True
    >>> p.preview
    1
    >>> p.isFirst()
    True
    >>> p.peek(3)
    [1, 2, 3]
    >>> p.next(2)
    [1, 2]
    >>> p.peek(2) #doctest: +ELLIPSIS
    [3, <object object at ...>]
    >>> p.peek(4) #doctest: +ELLIPSIS
    [3, <object object at ...>, <object object at ...>, <object object at ...>]
    >>> p.peek(1)
    [3]
    >>> p.next(2)
    Traceback (most recent call last):
    StopIteration
    >>> p.next()
    3
    >>> p.isLast()
    True
    >>> p.next()
    Traceback (most recent call last):
    StopIteration
    >>> p.next(0)
    []
    >>> p.peek()  #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.preview #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.isLast()  # after the iter process p.isLast remains True
    True

    From example 16.7 from python cookbook 2.

    The preview can be inspected through it.preview

    ignoring duplicates:
    >>> it = peek_ahead('122345567')
    >>> for i in it:
    ...     if it.preview == i:
    ...         continue
    ...     print(i, end=" ")
    1 2 3 4 5 6 7 

    getting duplicates together:
    >>> it = peek_ahead('abbcdddde')
    >>> for i in it:
    ...     if it.preview == i:
    ...         dup = 1
    ...         while 1:
    ...             i = it.next()
    ...             dup += 1
    ...             if i != it.preview:
    ...                 print(i*dup, end=" ")
    ...                 break
    ...     else:
    ...         print(i, end=" ")
    ...
    a bb c dddd e 
    
    """
    ## schildwacht (guard)
    sentinel = object()
    def __init__(self, iterable):
        ## iterator
        self._iterable = iter(iterable)
        try:
           ## next method hold for speed
            self._nit = self._iterable.next
        except AttributeError:
            self._nit = self._iterable.__next__
        ## deque object initialized left-to-right (using append())
        self._cache = collections.deque()
        ## initialize the first preview already
        self._fillcache(1)
        ## peek at leftmost item
        self.preview = self._cache[0]
        ## keeping the count allows checking isFirst and isLast status
        self.count = -1

    def __iter__(self):
        """return an iterator
        """
        return self

    def _fillcache(self, n):
        """fill _cache of items to come, with one extra for the preview variable
        """
        if n is None:
            n = 1
        while len(self._cache) < n+1:
            try:
                Next = self._nit()
            except StopIteration:
                # store sentinel, to identify end of iter:
                Next = self.sentinel
            self._cache.append(Next)

    def __next__(self, n=None):
        """gives next item of the iter, or a list of n items
        
        raises StopIteration if the iter is exhausted (self.sentinel is found),
        but in case of n > 1 keeps the iter alive for a smaller "next" calls
        """
        self._fillcache(n)
        if n is None:
            result = self._cache.popleft()
            if result == self.sentinel:
                # find sentinel, so end of iter:
                self.preview = self._cache[0]
                raise StopIteration
            self.count += 1
        else:
            result = [self._cache.popleft() for i in range(n)]
            if result and result[-1] == self.sentinel:
                # recache for future use:
                self._cache.clear()
                self._cache.extend(result)
                self.preview = self._cache[0]
                raise StopIteration
            self.count += n
        self.preview = self._cache[0]
        return result

    def next(self,n=None):
        """python2 compatibility
        """
        return self.__next__(n)
    
    def isFirst(self):
        """returns true if iter is at first position
        """
        return self.count == 0

    def isLast(self):
        """returns true if iter is at last position or after StopIteration
        """
        return self.preview == self.sentinel

    def hasNext(self):
        """returns true if iter is not at last position
        """
        return not self.isLast()
        
    def peek(self, n=None):
        """gives next item, without exhausting the iter, or a list of 0 or more next items
        
        with n == None, you can also use the self.preview variable, which is the first item
        to come.
        """
        self._fillcache(n)
        if n is None:
            result = self._cache[0]
        else:
            result = [self._cache[i] for i in range(n)]
        return result
    # old name:
    # peekable = peek_ahead

## TODOQH
class peek_ahead_stripped(peek_ahead):
    """ Iterator that strips lines of text, and returns (leftSpaces,strippedLine)

    sentinel is just False, such that peeking ahead can check for truth input

    >>> lines = ['line1', '', ' one space ahead','', '   three spaces ahead, 1 empty line before']
    >>> list(peek_ahead_stripped(lines))
    [(0, 'line1'), (0, ''), (1, 'one space ahead'), (0, ''), (3, 'three spaces ahead, 1 empty line before')]

    example of testing look ahead

    >>> lines = ['line1 ', '', 'line2 (last)']
    >>> it = peek_ahead_stripped(lines)
    >>> for spaces, text in it:
    ...     print('current line: |', text, '|', end=' ')
    ...     if it.preview is it.sentinel:
    ...         print(', cannot preview, end of peek_ahead_stripped')
    ...     elif it.preview[1]:
    ...         print(', non empty preview: |', it.preview[1], '|')
    ...     else:
    ...         print(', empty preview')
    current line: | line1 | , empty preview
    current line: |  | , non empty preview: | line2 (last) |
    current line: | line2 (last) | , cannot preview, end of peek_ahead_stripped

    """
    def _fillcache(self, n):
        """fill _cache of items to come, special treatment for this stripped subclass
        """
        if n is None:
            n = 1
        while len(self._cache) < n+1:
            try:
                line = self._nit()
                line = line.rstrip()
                Next = (len(line) - len(line.lstrip()), line.lstrip())
            except StopIteration:
                # store sentinel, to identify end of iter:
                Next = self.sentinel
            self._cache.append(Next)
     

def isSubList(largerList, smallerList):
    """returns 1 if smallerList is a sub list of largerList

>>> isSubList([1,2,4,3,2,3], [2,3])
True
>>> isSubList([1,2,3,2,2,2,2], [2])
True
>>> isSubList([1,2,3,2], [2,4])
False
    """
    if not smallerList:
        raise ValueError("isSubList: smallerList is empty: %s"% smallerList)
    item0 = smallerList[0]
    lenSmaller = len(smallerList)
    lenLarger = len(largerList)
    if lenSmaller > lenLarger:
        return False  # can not be sublist
    # get possible relevant indexes for first item
    indexes0 = [i for (i,item) in enumerate(largerList) if item == item0 and i <= lenLarger-lenSmaller]
    if not indexes0:
        return False
    for start in indexes0:
        _slice = largerList[start:start+lenSmaller]
        if _slice == smallerList:
            return True
    return False

# helper string functions:
def replaceExt(fileName, ext):
    """change extension of file

>>> replaceExt("a.psd", ".jpg")
'a.jpg'
>>> replaceExt("a/b/c/d.psd", "jpg")
'a/b/c/d.jpg'
    """
    ext = addToStart(ext, ".")
    fileName = str(fileName)
    a, _extOld = os.path.splitext(fileName)
    return a + ext

def getExt(fileName):
    """return the extension of a file

>>> getExt(u"a.psd")
'.psd'
>>> getExt("a/b/c/d.psd")
'.psd'
>>> getExt("abcd")
''
>>> getExt("a/b/xyz")
''
    """
    _a, ext = os.path.splitext(fileName)
    return str(ext)

def fileHasImageExtension(fileName):
    """return True if fileName has extension .jpg, .jpeg or .png
>>> fileHasImageExtension(u"a.JPG")
True
>>> fileHasImageExtension(u"yyy.JPEG")
True
>>> fileHasImageExtension(u"C:/a/b/d/e/xxx.png")
True
>>> fileHasImageExtension(u"a.txt")
False

    """
    ext = getExt(fileName)
    if not ext:
        return None
    return ext.lower() in [".jpg", ".jpeg", ".png"]

def fileHasJpgExtension(fileName):
    """return True if fileName has extension .jpg, .jpeg
>>> fileHasJpgExtension(u"a.JPG")
True
>>> fileHasJpgExtension(u"yyy.PNG")
False

    """
    ext = getExt(fileName)
    if not ext:
        return None
    return ext.lower() in [".jpg", ".jpeg"]

def removeFromStart(text, toRemove, ignoreCase=None):
    """returns the text with "toRemove" stripped from the start if it matches
>>> removeFromStart('abcd', 'a')
'bcd'
>>> removeFromStart('abcd', 'not')
'abcd'

working of ignoreCase:

>>> removeFromStart('ABCD', 'a')
'ABCD'
>>> removeFromStart('ABCD', 'ab', ignoreCase=1)
'CD'
>>> removeFromStart('abcd', 'ABC', ignoreCase=1)
'd'

    """
    if ignoreCase:
        text2 = text.lower()
        toRemove = toRemove.lower()
    else:
        text2 = text
    if text2.startswith(toRemove):
        return text[len(toRemove):]
    return text

def removeFromEnd(text, toRemove, ignoreCase=None):
    """returns the text with "toRemove" stripped from the end if it matches

>>> removeFromEnd('a.jpg', '.jpg')
'a'
>>> removeFromEnd('b.jpg', '.gif')
'b.jpg'

working of ignoreCase:

>>> removeFromEnd('C.JPG', '.jpg')
'C.JPG'
>>> removeFromEnd('D.JPG', '.jpg', ignoreCase=1)
'D'
>>> removeFromEnd('d.jpg', '.JPG', ignoreCase=1)
'd'

    """
    if ignoreCase:
        text2 = text.lower()
        toRemove = toRemove.lower()
    else:
        text2 = text
    if text2.endswith(toRemove):
        return text[:-len(toRemove)]
    return text

def addToStart(text, toAdd, ignoreCase=None):
    """returns text with "toAdd" added at the start if it was not already there

    if ignoreCase:
        return the start of the string with the case as in "toAdd"

>>> addToStart('a-text', 'a-')
'a-text'
>>> addToStart('text', 'b-')
'b-text'
>>> addToStart('B-text', 'b-')
'b-B-text'

working of ignoreCase:

>>> addToStart('C-Text', 'c-', ignoreCase=1)
'c-Text'
>>> addToStart('d-Text', 'D-', ignoreCase=1)
'D-Text'

    """
    if ignoreCase:
        text2 = text.lower()
        toAdd2 = toAdd.lower()
    else:
        text2 = text
        toAdd2 = toAdd
    if text2.startswith(toAdd2):
        return toAdd + text[len(toAdd):]
    return toAdd + text

def addToEnd(text, toAdd, ignoreCase=None):
    """returns text with "toAdd" added at the end if it was not already there

    if ignoreCase:
        return the end of the string with the case as in "toAdd"

>>> addToEnd('a.jpg', '.jpg')
'a.jpg'
>>> addToEnd('b', '.jpg')
'b.jpg'

working of ignoreCase:

>>> addToEnd('Cd.JPG', '.jpg', ignoreCase=1)
'Cd.jpg'
>>> addToEnd('Ef.jpg', '.JPG', ignoreCase=1)
'Ef.JPG'

    """
    if ignoreCase:
        text2 = text.lower()
        toAdd2 = toAdd.lower()
    else:
        text2 = text
        toAdd2 = toAdd
    if text2.endswith(toAdd2):
        return text[:-len(toAdd)] + toAdd
    return text + toAdd

def firstLetterCapitalize(t):
    """capitalize only the first letter of the string
    """
    if t:
        return t[0].upper() + t[1:]
    return ""

def extToLower(fileName):
    """leave name part intact, but change extension to lowercase
>>> extToLower("aBc.jpg")
'aBc.jpg'
>>> extToLower("ABC.JPG")
'ABC.jpg'
>>> extToLower("D:/a/B/ABC.JPG")
'D:/a/B/ABC.jpg'



    """
    f, ext = os.path.splitext(fileName)
    return f + ext.lower()


def appendBeforeExt(text, toAppend):
    """append text just before the extension of the filename part

>>> appendBeforeExt("short.html", '__t')
'short__t.html'
>>> appendBeforeExt("http://a/b/c/d/long.html", '__b')
'http://a/b/c/d/long__b.html'
    """
    base, ext = os.path.splitext(text)
    return base + toAppend + ext

def getBaseFolder(globalsDict):
    """get in a module the folder of this module.

    either sys.argv[0] (when run direct) or
    __file__, which can be empty. In that case take the working directory
    """
    baseFolder = ""
    if globalsDict['__name__']  == "__main__":
        baseFolder = os.path.split(sys.argv[0])[0]
        print('baseFolder from argv: %s'% baseFolder)
    elif globalsDict['__file__']:
        baseFolder = os.path.split(globalsDict['__file__'])[0]
        print('baseFolder from __file__: %s'% baseFolder)
    if not baseFolder or baseFolder == '.':
        baseFolder = os.getcwd()
    return baseFolder

Classes = ('ExploreWClass', 'CabinetWClass')


def partition_range(max_range):
    """partition for milla website in lengths of 3 or 4
    
>>> partition_range(3)
[[0], [1], [2]]
>>> partition_range(5)
[[0], [1], [2], [3, 4]]
>>> partition_range(8)
[[0, 1], [2, 3], [4, 5], [6, 7]]
>>> partition_range(12)
[[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]]

    """
    lst = list(range(max_range))
    if max_range <= 4:
        return [[i] for i in lst]
    if max_range == 5:
        L = [[i] for i in range(4)]
        L[3].append(4)
        return L
    if max_range == 6:
        return [[0, 1], [2, 3], [4, 5]]
    if max_range == 7:
        return [[0], [1, 2], [3, 4], [5, 6]]
    if max_range == 8:
        return [[0, 1], [2, 3], [4, 5], [6, 7]]
    return [lst] # alle images achter elkaar, scrollen afhankelijk van de browser
    



        

def unravelList(menu):
    """unravel from menu list to dropdown list order
    
>>> unravelList([1, [2, [3, 4, 5], 6], 7, 8])
[[1, 7, 8], [2, 6], [3, 4, 5]]

    """
    L, M = [], None
    for elt in menu:
        if isinstance(elt, list):
            M = unravelList(elt)
        else:
            L.append(elt)
    if M and isinstance(M, list):
        M.insert(0, L)
        return M
    return [L]

## to pathqh:
# # def toUnixName(t, glueChar="", lowercase=1, canHaveExtension=1, canHaveFolders=1, mayBeEmpty=False):
    
def sendkeys_escape(Str):
    """escape with {} keys that have a special meaning in sendkeys
    + ^ % ~ { } [ ]

>>> sendkeys_escape('abcd')
'abcd'
>>> sendkeys_escape('+bcd')
'{+}bcd'
>>> sendkeys_escape('+a^b%c~d{f}g[h]i')
'{+}a{^}b{%}c{~}d{{}f{}}g{[}h{]}i'
>>> sendkeys_escape('+^%~{}[]')
'{+}{^}{%}{~}{{}{}}{[}{]}'

    """
    ## make str, for python 2 for the time being:
    return ''.join(map(_sendkeys_escape, Str))

def _sendkeys_escape(s):
    """Escape one character in the set or return if different"""
    if s in ('+', '^', '%', '~', '{' , '}' , '[' , ']' ) :
        result = '{%s}' % s
    else:
        result = s
    return str(result)

def print_exc_plus(filename=None, skiptypes=None, takemodules=None,
                   specials=None):
    """ Print the usual traceback information, followed by a listing of
    all the local variables in each frame.
    
    
    """
    #print('specials:', specials')
    # normal traceback:
    traceback.print_exc()
    tb = sys.exc_info()[2]
    while tb.tb_next:
        tb = tb.tb_next
    stack = []
    f = tb.tb_frame
    while f:
        stack.append(f)
        f = f.f_back
    stack.reverse()
    traceback.print_exc()
    L = []
    # keys that are in specialsSitegen are recorded in next array:
    specialsDict = {}
    push = L.append

    push('traceback date/time: %s'% time.asctime(time.localtime(time.time())))
    pagename = ''
    menuname = ''
    for frame in stack:
        if takemodules and not [_f for _f in [frame.f_code.co_filename.find(t) > 0 for t in takemodules] if _f]:
            continue
        functionname = frame.f_code.co_name
        push('\nFrame "%s" in %s at line %s' % (frame.f_code.co_name,
                                frame.f_code.co_filename,
                                frame.f_lineno))
        keys = []
        values = []
        for key, value in list(frame.f_locals.items()):
            if key[0:2] == '__':
                continue
            try:
                v = repr(value)
            except:
                continue
            if skiptypes and [_f for _f in [v.find(s) == 1 for s in skiptypes] if _f]:
                continue
            keys.append(key)
            if functionname == 'go' and key == 'self':
                if v.find('Menu instance') > 0:
                    menuname = value.name
                    push('menu name: %s'% menuname)
            if functionname == 'makePage' and key == 'self':
                if v.find('Page instance') > 0:
                    pagename = value.name
                    push('page name: %s'% pagename)

            # we must _absolutely_ avoid propagating exceptions, and value
            # COULD cause any exception, so we MUST catch any...:
            v = v.replace('\n', '|')
            values.append(v)
        if keys:
            maxlenkeys = max(15, max(list(map(len, keys))))
            allowedlength = 80-maxlenkeys
            kv = list(zip(keys, values))
            kv.sort()
            for k,v in kv:
                if v.startswith('<built-in method'):
                    continue
                if len(v) > allowedlength:
                    half = allowedlength//2
                    v = v[:half] + " ... " + v[-half:]
                push(k.rjust(maxlenkeys) + " = " + v)
    if specials:
        stack.reverse()
        for frame in stack:
            if 'self' in frame.f_locals:
                push('\ncontents of self (%s)'% repr(frame.f_locals['self']))
                inst = frame.f_locals['self']
                keys, values = [], []
                for key in dir(inst):
                    value = getattr(inst, key)
                    if key[0:2] == '__':
                        continue
                    try:
                        v = repr(value)
                    except:
                        continue
                    if skiptypes and [_f for _f in [v.find(s) == 1 for s in skiptypes] if _f]:
                        continue
                    # specials for eg sitegen
                    if specials and key in specials:
                        #print('found specialskey: %s: %s'% (key, v)')
                        specialsDict[key] = v
                    keys.append(key)
                    # we must _absolutely_ avoid propagating exceptions, and value
                    # COULD cause any exception, so we MUST catch any...:
                    v = v.replace('\n', '|')
                    values.append(v)
                if not keys:
                    break
                maxlenkeys = max(15, max(list(map(len, keys))))
                allowedlength = 80-maxlenkeys
                for k,v in zip(keys, values):
                    if v and len(v) > allowedlength:
                        half = allowedlength//2
                        v = v[:half] + " ... " + v[-half:]
                    push(k.rjust(maxlenkeys) + " = " + str(v))
                break
            print('no self of HTMLDoc found')

    callback = []

    if menuname:
        push('menu: %s'% menuname)
        callback.append('menu: %s'% menuname)
    elif pagename == 'index':
        push('menu: top')
        callback.append('menu: top')
        
    if pagename:
        push('page: %s'% pagename)
        callback.append('page: %s'% pagename)
    push('\ntype: %s, value: %s\n'% (sys.exc_info()[0], sys.exc_info()[1]))
    callback.append('error: %s'%sys.exc_info()[1])

    print('\nerror occurred:')
    callback = '\n'.join(callback)
    print(callback)

    sys.stderr.write('\n'.join(L))
    sys.stderr.write(callback)
    if filename:
        print("skip writing to file %s"% filename)
        # with open(filename, 'w') as fout:
        #     fout.write("\n".join(L))
        #     print("written traceback in %s"% filename)    
    #print('result specialsDict: %s'% specialsDict')
    return callback, specialsDict

def cleanTraceback(tb, filesToSkip=None):
    """strip boilerplate in traceback (unittest)

    the purpose is to skip the lines "Traceback" (only if filesToSkip == True),
    and to skip traceback lines from modules that are in filesToSkip.

    in use with unimacro unittest and voicecode unittesting.

    filesToSkip are (can be "unittest.py" and "TestCaseWithHelpers.py"

    """
    L = tb.split('\n')
    snip = "  ..." # leaving a sign of the stripping!
    if filesToSkip:
        singleLineSkipping = ["Traceback (most recent call last):"]
    else:
        singleLineSkipping = None
    M = []
    skipNext = 0
    for line in L:
        # skip the traceback line:
        if singleLineSkipping and line in singleLineSkipping:
            continue
        # skip trace lines from one one the filesToSkip and the next one
        # UNLESS there are no leading spaces, in case we hit on the error line itself.
        if skipNext and line.startswith(" "):
            skipNext = 0
            continue
        if filesToSkip:
            for f in filesToSkip:
                if line.find(f + '", line') >= 0:
                    skipNext = 1
                    if M and  M[-1] == snip:
                        pass
                    else:
                        M.append(snip)
                    break
            else:
                skipNext = 0
                M.append(line)

    return '\n'.join(M)

def getSublists(L, maxLen, sepLen):
    """generator function, that gives pieces of the list, up to
    the maximum length, accounting for the separator length
    """
    if not L:
        yield L
        return
    listPart = [L[0]]
    lenPart = len(L[0])
    for w in L[1:]:
        lw = len(w)
        if lw + lenPart > maxLen:
            yield listPart
            listPart = [w]
            lenPart = lw
        else:
            lenPart += lw + sepLen
            listPart.append(w)
    yield listPart

def getWordsUntilLength(t, maxLength):
    """take words until maxLength is reached
>>> getWordsUntilLength('this is a test', 60)
'this is a test'
>>> getWordsUntilLength('this is a test', 7)
'this is'
>>> getWordsUntilLength('this is a test', 2)
'this'
    
    
    """
    t = t.replace(',', '')
    t = t.replace('.', '')
    T = t.split()
    while T:
        t = ' '.join(T)
        if len(t) <= maxLength:
            return t
        T.pop()
    return t

def splitLongString(S, maxLen=70, prefix='', prefixOnlyFirstLine=0):
    """Splits a (long) string into newline separated parts,

    a list of strings is returned.

    possibly with a fixed prefix, or a prefix for the first line only.
    Possibly items inside the line are separated by a given separator

    maxLen = maximum line length, can be exceeded is a very long word is there
    prefix = text that is inserted in front of each line, default ''
    prefixOnlyFirstLine = 1: following lines as blank prefix, default 0
    >>> splitLongString('foo', 80)
    ['foo']
    >>> splitLongString(' foo   bar and another set of  words  ', 80)
    ['foo bar and another set of words']
    >>> splitLongString(' foo   bar and another set of  words  ', 20,
    ... prefix='    # ')
    ['    # foo bar and', '    # another set of', '    # words']
    >>> splitLongString(' foo   bar and another set of  words  ', 20,
    ... prefix='entry = ', prefixOnlyFirstLine=1)
    ['entry = foo bar and', '        another set', '        of words']
    """
    assert isinstance(S, str)
    L = [t.strip() for t in S.split()]
    lOut = []
    for part in getSublists(L, maxLen=maxLen-len(prefix), sepLen=1):
        lOut.append(prefix + ' '.join(part))
        if prefixOnlyFirstLine:
            prefix = ' '*len(prefix)
    return lOut



def cleanString(s):
    """converts a string with leading and trailing and
    intermittent whitespace into a string that is stripped
    and has only single spaces between words
>>> cleanString('foo bar')
'foo bar'
>>> cleanString('foo  bar')
'foo bar'
>>> cleanString('\\n foo \\n\\n  bar ')
'foo bar'
>>> cleanString('')
''

    """
    return ' '.join([x.strip() for x in s.split()])

def formatListColumns(List, lineLen = 70, sort = 0):
    """formats a list in columns

    Uses a generator function "splitList", that gives a sequence of
    sub lists of length n.

    The items are separated by at least two spaces, if the list
    can be placed on one line, the list is comma separated

>>> formatListColumns([''])
''
>>> formatListColumns(['a','b'])
'a, b'
>>> formatListColumns(['foo', 'bar', 'longer entry'], lineLen=5)
'foo\\nbar\\nlonger entry'
>>> formatListColumns(['foo', 'bar', 'longer entry'], lineLen=5, sort=1)
'bar\\nfoo\\nlonger entry'
>>> print(formatListColumns(['afoo', 'bar', 'clonger', 'dmore', 'else', 'ftest'], lineLen=20, sort=1))
afoo     dmore
bar      else
clonger  ftest
>>> print(formatListColumns(['foo', 'bar', 'longer entry'], lineLen=20))
foo  longer entry
bar

    """
    if sort:
        List = sorted(List, key=str.casefold)
    s = ', '.join(List)

    # short list, simply join with comma space:
    if len(s) <= lineLen:
        return s

    maxLen = max(list(map(len, List)))

    # too long elements in list, return "\n" separated string:
    if maxLen > lineLen:
        return '\n'.join(List)


    nRow = len(s)//lineLen + 1
    lenList = len(List)

    # try for successive number of rows:
    while nRow < lenList//2 + 2:
        lines = []
        for i in range(nRow):
            lines.append([])
        maxLenTotal = 0
        for parts in splitList(List, nRow):
            maxLenParts = max(list(map(len, parts))) + 2
            maxLenTotal += maxLenParts
            for i, part in enumerate(parts):
                lines[i].append(part.ljust(maxLenParts))
        if maxLenTotal > lineLen:
            nRow += 1
        else:
            # return '\n'.join(map(string.strip, map(string.join, lines)))
            return '\n'.join([''.join(t).strip() for t in lines])
    # unexpected long list:
    return '\n'.join(List)

hasDoubleQuotes = re.compile(r'^".*"$')
hasSingleQuotes = re.compile(r"^'.*'$")

def convertToPythonArgs(text):
    """convert to numbers and strings

    IF argument is enclosed in " " or ' ' it is kept as a string.
    """    
    text = text.strip()
    if not text:
        return None
    L = text.split(',')
    L = list(map(_convertToPythonArg, L))
    return tuple(L)

def convertToPythonArgsKwargs(text):
    """convert to numbers and strings,

    IF argument is enclosed in " " or ' ' it is kept as a string.

    also do kwargs now...
>>> convertToPythonArgsKwargs('')
((), {})
>>> convertToPythonArgsKwargs('hello')
(('hello',), {})
>>> convertToPythonArgsKwargs('width=50')
((), {'width': 50})
>>> convertToPythonArgsKwargs('"hello", width=50')
(('hello',), {'width': 50})

    

    """
    L = []
    K = {}
    text = text.strip()
    if not text:
        return tuple(L), K 
    textList = text.split(',')
    for arg in textList:
        if arg.find("=") > 0:
            kw, arg2 = [t.strip() for t in arg.split("=", 1)]
            if kw.find(" ") == -1:
                arg2 = _convertToPythonArg(arg2)
                K[kw] = arg2
            else:
                L.append(arg.strip())
        else:
            L.append(_convertToPythonArg(arg))
    return tuple(L), K

def _convertToPythonArg(t):
    t = t.strip()
    if not t:
        return ''

    # if input string is a number, return string directly
    try:
        i = int(t)
        if t == '0':
            return 0
        if t.startswith('0'):
            print('warning convertToPythonArg, could be int, but assume string: %s'% t)
            return '%s'% t
        return i
    except ValueError:
        pass
    try:
        f = float(t)
        if t.find(".") >= 0:
            return f
        print('warning convertToPythonArg, can be float, but assume string: %s'% t)
        return '%s'% t
    except ValueError:
        pass

    # now proceeding with strings:    
    if hasDoubleQuotes.match(t):
        return t[1:-1]
    if hasSingleQuotes.match(t):
        return t[1:-1]
    return t

def splitList(L, n):
    """generator function that splits a list in sublists of length n

    """
    O = []
    for l in L:
        O.append(l)
        if len(O) == n:
            yield O
            O = []
    if O:
        yield O

def makeReadable(t):
    """squeeze text for readability
    
    helper for print lines...
    """
    t = t.strip()
    t = t.replace('\n', '\\\\')
    if len(t) > 100:
        return t[:50] + ' ... ' + t[-50:]
    return t

## functions for generating alternative paths in virtual drives
## uses reAltenativePaths, defined in the top of this module
## put in utilsqh.py! used in sitegen AND in _folders.py grammar of Unimacro:
# for alternatives in virtual drive definitions:
reAltenativePaths = re.compile(r"(\([^|()]+?(\|[^|()]+?)+\))")

def generate_alternatives(s):
    """generates altenatives if (xxx|yyy) is found, otherwise just yields s
    Helper for cross_loop_alternatives
    """
    m = reAltenativePaths.match(s)
    if m:
        alternatives = s[1:-1].split("|")
        for item in alternatives:
            yield item
    else:
        yield s
        
def cross_loop_alternatives(*sequences):
    """helper function for loop_through_alternative_paths
    """
    if sequences:
        for x in generate_alternatives(sequences[0]):
            for y in cross_loop_alternatives(*sequences[1:]):
                yield (x,) + y
    else:
        yield ()

def loop_through_alternative_paths(pathdefinition):
    """can hold alternatives (a|b)
    
>>> list(loop_through_alternative_paths("(C|D):/xxxx/yyyy"))
['C:/xxxx/yyyy', 'D:/xxxx/yyyy']
>>> list(loop_through_alternative_paths("(C:|D:|E:)/Do(kum|cum)ent(s|en)"))
['C:/Dokuments', 'C:/Dokumenten', 'C:/Documents', 'C:/Documenten', 'D:/Dokuments', 'D:/Dokumenten', 'D:/Documents', 'D:/Documenten', 'E:/Dokuments', 'E:/Dokumenten', 'E:/Documents', 'E:/Documenten']

    So "(C|D):/natlink" first yields "C:/natlink" and then "D:/natlink".
    More alternatives in one item are possible, see second example.
    """
    m = reAltenativePaths.search(pathdefinition)
    if m:
        result = reAltenativePaths.split(pathdefinition)
        result = [x for x in result if x and not x.startswith("|")]
        for pathdef in cross_loop_alternatives(*result):
            yield ''.join(pathdef)
    else:
        # no alternatives, simply yield the pathdefinition:
        yield pathdefinition

def getValidPath(variablePathDefinition):
    r"""check the different alternatives of the definition
    
# >>> homefolder = getValidPath(r'(A|B|D|C|D):\Users\(Dell|Gebruiker)\.Natlink')
# >>> homefolder
# WindowsPath('C:/Users/Gebruiker/.Natlink')
# >>> str(homefolder)
# 'C:\\Users\\Gebruiker\\.Natlink'

    return the first valid path, None if not found
    """
    for p in loop_through_alternative_paths(variablePathDefinition):
        if os.path.exists(p):
            return Path(p)
    return None

def _test():
    #pylint:disable=C0415
    import doctest
    return doctest.testmod()

if __name__ == "__main__":


    def revTrue(t):
        return "y"
    def revFalse(t):
        return "n"
    def revAbort(t):
        return
    _test()



"""actions from application pythonwin
"""
import win32gui
import win32con
import unimacro.messagefunctions
import as
import mf
from .actionbases import MessageActions
import pprint
class PythonwinActions(MessageActions):
    handlesDict = {}
    def __init__(self, progInfo):
        MessageActions.__init__(self, progInfo)
        
    def getInnerHandle(self, handle):
        
        controls = mf.findControls(handle, wantedClass="Scintilla")
        print 'Scintilla controls: %s'% controls
        for c in controls:
            ln = self.getCurrentLineNumber(c)
            numberLines = self.getNumberOfLines(c)
            visible1 = mf.isVisible(c)
            print 'c: %s, linenumber: %s, nl: %s, visible: %s'% (c, ln, numberLines, visible1)
                
if __name__ == '__main__':

    pass